# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, GLib

from lollypop.define import App, Type, ViewType, StorageType
from lollypop.utils import emit_signal, get_default_storage_type


class ViewsContainer:
    """
        Views management for main view
    """

    def __init__(self):
        """
            Init container
        """
        pass

    def show_widget(self, widget):
        """
            Show widget
            @param widget as Gtk.Widget
        """
        from lollypop.view import View
        view = View(StorageType.ALL, ViewType.DEFAULT)
        view.show()
        view.add(widget)
        widget.set_vexpand(True)
        self._stack.add(view)
        self._stack.set_visible_child(view)

    def show_menu(self, widget):
        """
            Show menu widget
            @param widget as Gtk.Widget
        """
        def on_hidden(widget, hide, view):
            if hide:
                self._main_stack.set_transition_type(
                    Gtk.StackTransitionType.SLIDE_UP)
                self._main_stack.set_visible_child_name("main")
                menu = self._main_stack.get_child_by_name("menu")
                GLib.timeout_add(500, self._main_stack.remove, menu)
                self._main_stack.set_transition_type(
                    Gtk.StackTransitionType.CROSSFADE)
                App().enable_special_shortcuts(True)
            if self.can_go_back:
                emit_signal(self, "can-go-back-changed", True)

        from lollypop.view_menu import MenuView
        view = MenuView(widget)
        view.show()
        widget.connect("hidden", on_hidden, view)
        self._main_stack.add_named(view, "menu")
        self._main_stack.set_transition_type(
            Gtk.StackTransitionType.SLIDE_DOWN)
        self._main_stack.set_visible_child(view)
        self._main_stack.set_transition_type(
            Gtk.StackTransitionType.CROSSFADE)
        emit_signal(self, "can-go-back-changed", False)
        App().enable_special_shortcuts(False)
        self.dismiss_notification()

    def show_view(self, item_ids, data=None, storage_type=None):
        """
            Show view for item id
            @param item_ids as [int]
            @param data as object
            @param storage_type as StorageType
        """
        self.dismiss_notification()
        self.sub_widget.set_visible_child(self.grid_view)
        view = None
        if storage_type is None:
            storage_type = get_default_storage_type()
        hide_selection_list = True
        if item_ids:
            if item_ids[0] in [Type.POPULARS,
                               Type.LOVED,
                               Type.RECENTS,
                               Type.LITTLE,
                               Type.RANDOMS]:
                view = self._get_view_albums(item_ids, [], storage_type)
            elif item_ids[0] == Type.WEB:
                view = self._get_view_albums(item_ids, [], StorageType.SAVED)
            elif item_ids[0] == Type.SUGGESTIONS:
                view = self._get_view_suggestions(storage_type)
            elif item_ids[0] == Type.SEARCH:
                view = self.get_view_search(data)
            elif item_ids[0] == Type.INFO:
                view = self._get_view_info()
            elif item_ids[0] == Type.DEVICE_ALBUMS:
                view = self._get_view_device_albums(data)
            elif item_ids[0] == Type.DEVICE_PLAYLISTS:
                view = self._get_view_device_playlists(data)
            elif item_ids[0] == Type.LYRICS:
                view = self._get_view_lyrics()
            elif item_ids[0] == Type.GENRES:
                if data is None:
                    view = self._get_view_genres(storage_type)
                else:
                    view = self._get_view_albums([data], [], storage_type)
            elif item_ids[0] == Type.ALBUM:
                hide_selection_list = False
                view = self._get_view_album(data, storage_type)
            elif item_ids[0] == Type.YEARS:
                if data is None:
                    view = self._get_view_albums_decades(storage_type)
                else:
                    view = self._get_view_albums_years(data, storage_type)
            elif item_ids[0] == Type.PLAYLISTS:
                view = self._get_view_playlists(data)
            elif item_ids[0] == Type.EQUALIZER:
                from lollypop.view_equalizer import EqualizerView
                view = EqualizerView()
            elif item_ids[0] == Type.ALL:
                view = self._get_view_albums(item_ids, [], storage_type)
            elif item_ids[0] == Type.COMPILATIONS:
                view = self._get_view_albums(item_ids, [], storage_type)
            elif item_ids[0] == Type.ARTISTS:
                view = self._get_view_artists([], data, storage_type)
        self._sidebar.select_ids(item_ids, False)
        if hide_selection_list:
            self._hide_right_list()
            self.left_list.hide()
        if view is not None:
            self.set_focused_view(view)
            view.show()
            self._stack.add(view)
            self._stack.set_visible_child(view)
        emit_signal(self, "can-go-back-changed", self.can_go_back)

    def get_view_current(self):
        """
            Get view for current playlist
            @return View
        """
        from lollypop.view_current_albums import CurrentAlbumsView
        # Search view in children
        for (_view, _class, args, sidebar_id,
             selection_ids, position) in self._stack.history.items:
            if _class == CurrentAlbumsView and _view is not None:
                self._stack.history.remove(_view)
                return _view
        view = CurrentAlbumsView(ViewType.DND)
        view.populate()
        view.show()
        return view

    def get_view_search(self, search=""):
        """
            Get view for search
            @param search as str
            @return SearchView
        """
        from lollypop.view_search import SearchView
        view = None
        # Search view in current view
        if self.view is not None and isinstance(self.view, SearchView):
            view = self.view
        # Search view in children
        else:
            for (_view, _class, args, sidebar_id,
                 selection_ids, position) in self._stack.history.items:
                if _class == SearchView and _view is not None:
                    self._stack.history.remove(_view)
                    view = _view
                    break
        if view is None:
            view_type = ViewType.SCROLLED
            view = SearchView("", view_type)
            view.show()
        if search:
            view.set_search(search)
        else:
            view.grab_focus()
        return view

##############
# PROTECTED  #
##############
    def _get_view_playlists(self, playlist_id=None):
        """
            Get playlists view for playlists
            @param playlist_id as int
            @return View
        """
        view_type = ViewType.PLAYLISTS | ViewType.SCROLLED
        if playlist_id is None:
            from lollypop.view_playlists_manager import PlaylistsManagerView
            view = PlaylistsManagerView(view_type)
        else:
            from lollypop.view_playlists import PlaylistsView
            view_type |= ViewType.DND
            view = PlaylistsView(playlist_id, view_type)
        view.populate()
        return view

    def _get_view_device_playlists(self, index):
        """
            Show playlists for device at index
            @param index as int
        """
        view_type = ViewType.SCROLLED
        from lollypop.view_playlists_manager import PlaylistsManagerDeviceView
        view = PlaylistsManagerDeviceView(index, view_type)
        view.populate()
        return view

    def _get_view_lyrics(self):
        """
            Show lyrics for track
            @pram track as Track
        """
        view_type = ViewType.SCROLLED
        from lollypop.view_lyrics import LyricsView
        view = LyricsView(view_type)
        view.populate(App().player.current_track)
        view.show()
        return view

    def _get_view_artists_rounded(self, storage_type):
        """
            Get rounded artists view
            @param storage_type as StorageType
            @return RoundedArtistsViewWithBanner
        """
        view_type = ViewType.SCROLLED
        from lollypop.view_artists_rounded import RoundedArtistsViewWithBanner
        view = RoundedArtistsViewWithBanner(storage_type, view_type)
        self._stack.add(view)
        view.populate()
        view.show()
        return view

    def _get_view_artists(self, genre_ids, artist_ids, storage_type):
        """
            Get artists view for genres/artists
            @param genre_ids as [int]
            @param artist_ids as [int]
            @param storage_type as StorageType
            @return ArtistViewBox/ArtistViewList
        """
        view_type = ViewType.SCROLLED
        from lollypop.view_artist_list import ArtistViewList
        view = ArtistViewList(genre_ids, artist_ids,
                              storage_type, view_type)
        view.populate()
        view.show()
        return view

    def _get_view_suggestions(self, storage_type):
        """
            Get home view
            @param storage_type as StorageType
            @return SuggestionsView
        """
        view_type = ViewType.SCROLLED
        from lollypop.view_suggestions import SuggestionsView
        view = SuggestionsView(storage_type, view_type)
        view.populate()
        view.show()
        return view

    def _get_view_albums_decades(self, storage_type):
        """
            Get album view for decades
            @param storage_type as StorageType
            @return DecadesBoxView
        """
        view_type = ViewType.SCROLLED
        from lollypop.view_decades_box import DecadesBoxView
        view = DecadesBoxView(storage_type, view_type)
        view.populate()
        view.show()
        return view

    def _get_view_album(self, album, storage_type):
        """
            Show album
            @param album as Album
            @param storage_type as StorageType
            @return AlbumView
        """
        view_type = ViewType.SCROLLED |\
            ViewType.OVERLAY | ViewType.ALBUM
        from lollypop.view_album import AlbumView
        view = AlbumView(album, storage_type, view_type)
        view.populate()
        return view

    def _get_view_genres(self, storage_type):
        """
            Get view for genres
            @param storage_type as StorageType
            @return GenresBoxView
        """
        view_type = ViewType.SCROLLED
        from lollypop.view_genres_box import GenresBoxView
        view = GenresBoxView(storage_type, view_type)
        view.populate()
        view.show()
        return view

    def _get_view_albums_years(self, years, storage_type):
        """
            Get album view for years
            @param years as [int]
            @param storage_type as StorageType
            @return AlbumsForYearsBoxView
        """
        view_type = ViewType.SCROLLED
        from lollypop.view_albums_box import AlbumsForYearsBoxView
        view = AlbumsForYearsBoxView([Type.YEARS], years,
                                     storage_type, view_type)
        view.populate()
        return view

    def _get_view_albums(self, genre_ids, artist_ids, storage_type):
        """
            Get albums view for genres/artists
            @param genre_ids as [int]
            @param artist_ids as [int]
            @param storage_type as StorageType
            @return AlbumsForGenresBoxView
        """
        view_type = ViewType.SCROLLED
        from lollypop.view_albums_box import AlbumsForGenresBoxView
        view = AlbumsForGenresBoxView(genre_ids, artist_ids,
                                      storage_type, view_type)
        view.populate()
        return view

    def _get_view_device_albums(self, index):
        """
            Show albums for device at index
            @param index as int
            @return AlbumsDeviceBoxView
        """
        view_type = ViewType.SCROLLED
        from lollypop.view_albums_box import AlbumsDeviceBoxView
        view = AlbumsDeviceBoxView(index, view_type)
        view.populate()
        return view

    def _get_view_info(self):
        """
            Get view for information
            @return InformationView
        """
        from lollypop.view_information_stack import InformationViewStack
        view = InformationViewStack()
        view.populate()
        return view



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gio, GLib

from gettext import gettext as _

from lollypop.menu_sync import SyncAlbumsMenu
from lollypop.menu_playlists import PlaylistsMenu
from lollypop.define import App, ViewType, Type
from lollypop.objects_album import Album
from lollypop.utils import get_default_storage_type


class ArtistMenu(Gio.Menu):
    """
        Contextual menu for artist
    """
    def __init__(self, artist_id, storage_type, view_type, header=False):
        """
            Init artist menu
            @param artist_id as int
            @param storage_type as StorageType
            @param view_type as ViewType
            @param header as bool
        """
        Gio.Menu.__init__(self)
        if header:
            from lollypop.menu_header import ArtistMenuHeader
            self.append_item(ArtistMenuHeader(artist_id))
        if view_type & ViewType.BANNER:
            show_artist_tracks = App().settings.get_value("show-artist-tracks")
            action = Gio.SimpleAction.new_stateful(
                "show-artist-tracks",
                None,
                GLib.Variant.new_boolean(show_artist_tracks))
            App().add_action(action)
            action.connect("change-state", self.__on_change_state)
            self.append(_("Show tracks"), "app.show-artist-tracks")
            action = Gio.SimpleAction.new_stateful(
                "play-featured",
                None,
                GLib.Variant.new_boolean(
                    App().settings.get_value("play-featured")))
            App().add_action(action)
            action.connect("change-state", self.__on_change_state)
            self.append(_("Play featured"), "app.play-featured")
        else:
            from lollypop.menu_playback import ArtistPlaybackMenu
            self.append_section(_("Playback"),
                                ArtistPlaybackMenu(artist_id, storage_type))
        menu = Gio.Menu()
        self.append_section(_("Add to"), menu)
        storage_type = get_default_storage_type()
        album_ids = App().albums.get_ids([], [artist_id], storage_type, False)
        albums = [Album(album_id) for album_id in album_ids]
        menu.append_submenu(_("Devices"), SyncAlbumsMenu(albums))
        menu.append_submenu(_("Playlists"), PlaylistsMenu(albums))

#######################
# PRIVATE             #
#######################
    def __on_change_state(self, action, variant):
        """
            Save option and reload view
            @param Gio.SimpleAction
            @param GLib.Variant
        """
        name = action.get_name()
        action.set_state(variant)
        App().settings.set_value(name, variant)
        if name.find("show") != -1:
            App().window.container.reload_view()


class ArtistAlbumsMenu(Gio.Menu):
    """
        Contextual menu for artist albums
    """

    def __init__(self, artist_id, storage_type):
        """
            Init artist albums menu
            @param artist id as int
        """
        Gio.Menu.__init__(self)
        self.__storage_type = storage_type
        self.__artist_id = artist_id
        self.__set_actions()

#######################
# PRIVATE             #
#######################
    def __set_actions(self):
        """
            Set artist actions
        """
        go_artist_action = Gio.SimpleAction(name="go_artist_action")
        App().add_action(go_artist_action)
        go_artist_action.connect("activate",
                                 self.__on_go_to_artist_activate)
        self.append(_("Available albums"), "app.go_artist_action")

    def __on_go_to_artist_activate(self, action, variant):
        """
            Show albums from artists
            @param Gio.SimpleAction
            @param GLib.Variant
        """
        App().window.container.show_view([Type.ARTISTS],
                                         [self.__artist_id],
                                         self.__storage_type)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk

from gettext import gettext as _


class SmartPlaylistRow(Gtk.ListBoxRow):
    """
        A smart playlist widget (WHERE SQL subrequest)
    """
    __TEXT = ["genre", "album", "artist"]
    __INT = ["rating", "popularity", "year", "bpm"]

    def __init__(self, size_group):
        """
            Init widget
            @param size_group as Gtk.SizeGroup
        """
        Gtk.ListBoxRow.__init__(self)
        self.__operand = None
        builder = Gtk.Builder()
        builder.add_from_resource("/org/gnome/Lollypop/SmartPlaylistWidget.ui")
        builder.connect_signals(self)
        self.__stack = builder.get_object("stack")
        self.__type_combobox = builder.get_object("type_combobox")
        self.__operand_combobox = builder.get_object("operand_combobox")
        size_group.add_widget(self.__operand_combobox)
        self.__entry = builder.get_object("entry")
        self.__spin_button = builder.get_object("spin_button")
        self.__rate = 0
        self._stars = []
        self._stars.append(builder.get_object("star0"))
        self._stars.append(builder.get_object("star1"))
        self._stars.append(builder.get_object("star2"))
        self._stars.append(builder.get_object("star3"))
        self._stars.append(builder.get_object("star4"))
        self._on_leave_notify_event(None, None)
        self.add(builder.get_object("widget"))

    def set(self, item):
        """
            Try to set widget from item
            @param item as str
        """
        item = item.replace(" COLLATE NOCASE", "")
        # Sucks, search for a better way to handle this in Lollypop2
        if item.find("NOT LIKE") != -1:
            self.__operand = "NOT LIKE"
            (t, *args) = item.split(" NOT LIKE ")
        else:
            (t, self.__operand, *args) = item.split(" ")
        value = " ".join(list(args))
        # Unquote value
        if value[0] == "'":
            value = value[1:]
        if value[-1] == "'":
            value = value[:-1]
        # Remove %
        if value[0] == "%":
            value = value[1:]
        if value[-1] == "%":
            value = value[:-1]
        if t == "tracks.year":
            self.__type_combobox.set_active_id("year")
            self.__spin_button.set_value(int(value))
        elif t == "tracks.bpm":
            self.__type_combobox.set_active_id("bpm")
            self.__spin_button.set_value(int(value))
        elif t == "genres.name":
            self.__type_combobox.set_active_id("genre")
            self.__entry.set_text(value)
        elif t == "albums.name":
            self.__type_combobox.set_active_id("album")
            self.__entry.set_text(value)
        elif t == "artists.name":
            self.__type_combobox.set_active_id("artist")
            self.__entry.set_text(value)
        elif t == "tracks.rate":
            self.__type_combobox.set_active_id("rating")
            self.__rate = int(value)
            self._on_leave_notify_event(None, None)
        elif t == "tracks.popularity":
            self.__type_combobox.set_active_id("popularity")
            self.__rate = int(value)
            self._on_leave_notify_event(None, None)
        else:
            self.destroy()

    @property
    def sql(self):
        """
            Get SQL subrequest
            @return str
        """
        request_type = self.__type_combobox.get_active_id()
        request_check = self.__operand_combobox.get_active_id()
        sql = None
        if request_type == "rating":
            sql = "( ((tracks.rate %s '%s')) )"
            sql = sql % (request_check, self.__rate)
        elif request_type == "popularity":
            sql = "( ((tracks.popularity %s '%s')) )"
            sql = sql % (request_check, self.__rate)
        elif request_type == "year":
            value = int(self.__spin_button.get_value())
            sql = "( ((tracks.year %s '%s')) )"
            sql = sql % (request_check, value)
        elif request_type == "bpm":
            value = int(self.__spin_button.get_value())
            sql = "( ((tracks.bpm %s '%s')) )"
            sql = sql % (request_check, value)
        elif request_type == "genre":
            request_value = self.__entry.get_text().replace("'", "''")
            if request_check.find("LIKE") != -1:
                value = "%" + request_value + "%"
            else:
                value = request_value
            sql = "(album_genres.genre_id = genres.rowid" +\
                  " AND tracks.album_id = album_genres.album_id" +\
                  " AND ((genres.name %s '%s' COLLATE NOCASE)) )"
            sql = sql % (request_check, value)
        elif request_type == "album":
            request_value = self.__entry.get_text().replace("'", "''")
            if request_check.find("LIKE") != -1:
                value = "%" + request_value + "%"
            else:
                value = request_value
            sql = "(tracks.album_id = albums.rowid" +\
                  " AND ((albums.name %s '%s' COLLATE NOCASE)) )"
            sql = sql % (request_check, value)
        elif request_type == "artist":
            request_value = self.__entry.get_text().replace("'", "''")
            if request_check.find("LIKE") != -1:
                value = "%" + request_value + "%"
            else:
                value = request_value
            sql = "(track_artists.artist_id = artists.rowid" +\
                  " AND tracks.rowid = track_artists.track_id" +\
                  " AND ((artists.name %s '%s' COLLATE NOCASE)) )"
            sql = sql % (request_check, value)
        return sql

#######################
# PROTECTED           #
#######################
    def _on_destroy_button_clicked(self, button):
        """
            Destroy self
            @param button as Gtk.Button
        """
        self.destroy()

    def _on_type_combobox_changed(self, combobox):
        """
            Update check combobox
            @param combobox as Gtk.ComboBoxText
        """
        self.__operand_combobox.remove_all()
        self.__operand_combobox.append("=", _("is equal to"))
        self.__operand_combobox.append("!=", _("is not equal to"))
        if combobox.get_active_id() in self.__TEXT:
            self.__operand_combobox.append("LIKE", _("contains "))
            self.__operand_combobox.append("NOT LIKE", _("does not contain"))
        elif combobox.get_active_id() in self.__INT:
            self.__operand_combobox.append(">", _("is greater than"))
            self.__operand_combobox.append("<", _("is less than"))
        if self.__operand is None:
            self.__operand_combobox.set_active(0)
        else:
            self.__operand_combobox.set_active_id(self.__operand)

        active_id = combobox.get_active_id()
        if active_id in ["year", "bpm"]:
            self.__stack.set_visible_child_name("int")
            if active_id == "year":
                self.__spin_button.set_value(1960)
            else:
                self.__spin_button.set_value(130)
        elif combobox.get_active_id() in ["rating", "popularity"]:
            self.__stack.set_visible_child_name("rating")
        else:
            self.__stack.set_visible_child_name("text")

    def _on_enter_notify_event(self, widget, event):
        """
            Update star opacity
            @param widget as Gtk.EventBox
            @param event as Gdk.Event
        """
        event_star = widget.get_children()[0]
        # First star is hidden (used to clear score)
        if event_star.get_opacity() == 0.0:
            found = True
        else:
            found = False
        for star in self._stars:
            if found:
                star.set_opacity(0.2)
                star.get_style_context().remove_class("selected")
            else:
                star.get_style_context().add_class("selected")
                star.set_opacity(0.8)
            if star == event_star:
                found = True

    def _on_leave_notify_event(self, widget, event):
        """
            Update star opacity
            @param widget as Gtk.EventBox (can be None)
            @param event as Gdk.Event (can be None)
        """
        if self.__rate < 1:
            for i in range(5):
                self._stars[i].set_opacity(0.2)
                self._stars[i].get_style_context().remove_class("selected")
        else:
            star = self.__star_from_rate(self.__rate)
            # Select wanted star
            for idx in range(0, star):
                widget = self._stars[idx]
                widget.get_style_context().add_class("selected")
                widget.set_opacity(0.8)
            # Unselect others
            for idx in range(star, 5):
                self._stars[idx].set_opacity(0.2)
                self._stars[idx].get_style_context().remove_class("selected")

    def _on_button_release_event(self, widget, event):
        """
            Set album popularity
            @param widget as Gtk.EventBox
            @param event as Gdk.Event
        """
        event_star = widget.get_children()[0]
        if event_star in self._stars:
            position = self._stars.index(event_star)
        else:
            position = -1
        self.__rate = position + 1
        self._on_leave_notify_event(None, None)
        return True

#######################
# PRIVATE             #
#######################
    def __star_from_rate(self, rate):
        """
            Calculate stars from rate
            @param rate as double
            @return int
        """
        star = min(5, int(rate))
        return star



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import GLib, Gtk, Gdk

from gettext import gettext as _
from random import randint

from lollypop.define import App, NetworkAccessACL


class NotificationContainer:
    """
        Notification management for main view
    """

    __DONATION = 1 << 0
    __BACKGROUND_DATA = 1 << 1

    def __init__(self):
        """
            Init container
        """
        self.__notification = None
        notification_flag = App().settings.get_value(
            "notification-flag").get_int32()
        network_access_acl = App().settings.get_value(
            "network-access-acl").get_int32()
        if not notification_flag & self.__DONATION:
            GLib.timeout_add_seconds(randint(3600, 7200),
                                     self.__show_donation)
        if not notification_flag & self.__BACKGROUND_DATA and\
                not network_access_acl & NetworkAccessACL["DATA"]:
            self.__show_background_data()

    def show_notification(self, title, buttons, actions):
        """
            Show a notification
            @param title as str
            @param buttons as [str]
            @param actions as [function]
        """
        # Do not show multiple notifications
        if self.__notification is None:
            from lollypop.app_notification import AppNotification
            self.__notification = AppNotification(title, buttons, actions)
            self.__notification.connect("destroy",
                                        self.__on_notification_destroy)
            self.add_overlay(self.__notification)
            self.__notification.show()
            self.__notification.set_reveal_child(True)

    def dismiss_notification(self):
        """
            Get current notification
            @return AppNotification
        """
        if self.__notification is not None:
            self.__notification.destroy()

############
# PRIVATE  #
############
    def __set_gsettings_value(self, value):
        """
            Set notification flag value in Gsettings
            @param value as int
        """
        notification_flag = App().settings.get_value(
            "notification-flag").get_int32()
        notification_flag |= value
        App().settings.set_value("notification-flag",
                                 GLib.Variant("i", notification_flag))

    def __show_background_data(self):
        """
            Show a notification to enable background data
        """
        def enable_data():
            network_access_acl = App().settings.get_value(
                "network-access-acl").get_int32()
            network_access_acl |= NetworkAccessACL["DATA"]
            App().settings.set_value("network-access-acl",
                                     GLib.Variant("i", network_access_acl))

        self.show_notification(
            _("Automatically download albums and artists artwork?"),
            [_("Yes")],
            [enable_data])
        self.__set_gsettings_value(self.__BACKGROUND_DATA)

    def __show_donation(self):
        """
            Show a notification telling user to donate a little
        """
        self.show_notification(
            _("Please consider a donation to the project"),
            [_("PayPal"), _("Patreon")],
            [lambda: Gtk.show_uri_on_window(
                App().window,
                "https://www.paypal.me/lollypopgnome",
                Gdk.CURRENT_TIME),
             lambda: Gtk.show_uri_on_window(
                App().window,
                "https://www.patreon.com/gnumdk",
                Gdk.CURRENT_TIME)])
        self.__set_gsettings_value(self.__DONATION)

    def __on_notification_destroy(self, notification):
        self.__notification = None



# Copyright (c) 2014-2017 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import json

from lollypop.logger import Logger
from lollypop.define import App


class DeezerWebHelper:
    """
        Web helper for Deezer
    """

    def __init__(self):
        """
            Init helper
        """
        pass

    def get_artist_id(self, artist_name, cancellable):
        """
            Get artist id
            @param artist_name as str
            @param cancellable as Gio.Cancellable
            @return {}
        """
        try:
            uri = "https://api.deezer.com/search/artist?q=%s" % artist_name
            (status, data) = App().task_helper.load_uri_content_sync(
                uri, cancellable)
            if status:
                decode = json.loads(data.decode("utf-8"))
                for artist in decode["data"]:
                    return artist["id"]
        except Exception as e:
            Logger.warning("DeezerWebHelper::get_artist_id(): %s", e)
        return None

    def get_album_payload(self, album_id, cancellable):
        """
            Get album payload for id
            @param album_id as int
            @param cancellable as Gio.Cancellable
            @return {}
        """
        try:
            uri = "https://api.deezer.com/album/%s" % album_id
            (status, data) = App().task_helper.load_uri_content_sync(
                uri, cancellable)
            if status:
                decode = json.loads(data.decode("utf-8"))
                return decode
        except Exception as e:
            Logger.warning("DeezerWebHelper::get_album_payload(): %s", e)
        return None

    def get_track_payload(self, track_id, cancellable):
        """
            Get album payload for id
            @param album_id as int
            @param cancellable as Gio.Cancellable
            @return {}
        """
        try:
            uri = "https://api.deezer.com/track/%s" % track_id
            (status, data) = App().task_helper.load_uri_content_sync(
                uri, cancellable)
            if status:
                decode = json.loads(data.decode("utf-8"))
                return decode
        except Exception as e:
            Logger.warning("DeezerWebHelper::get_track_payload(): %s", e)
        return None

    def lollypop_album_payload(self, payload):
        """
            Convert payload to Lollypop one
            @param payload as {}
            return {}
        """
        lollypop_payload = {}
        lollypop_payload["mbid"] = None
        lollypop_payload["name"] = payload["title"]
        lollypop_payload["uri"] = "dz:%s" % payload["id"]
        lollypop_payload["artists"] = payload["artist"]["name"]
        lollypop_payload["track-count"] = payload["nb_tracks"]
        lollypop_payload["artwork-uri"] = payload["cover_big"]
        try:
            lollypop_payload["date"] = "%sT00:00:00" % payload["release_date"]
        except:
            lollypop_payload["date"] = None
        return lollypop_payload

    def lollypop_track_payload(self, payload):
        """
            Convert payload to Lollypop one
            @param payload as {}
            @return {}
        """
        lollypop_payload = {}
        lollypop_payload["mbid"] = None
        lollypop_payload["name"] = payload["title"]
        lollypop_payload["uri"] = "dz:%s" % payload["id"]
        lollypop_payload["artists"] = payload["artist"]["name"]
        try:
            lollypop_payload["discnumber"] = payload["disk_number"]
        except:
            lollypop_payload["discnumber"] = 1
        lollypop_payload["tracknumber"] = payload["track_position"]
        lollypop_payload["duration"] = payload["duration"] * 1000
        return lollypop_payload

#######################
# PRIVATE             #
#######################



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, GObject, GLib

from lollypop.utils import emit_signal
from lollypop.define import App


class ComboRow(Gtk.ListBoxRow):
    """
        A Row for combobox
    """

    def __init__(self, title):
        """
            Init widget
            @param title as str
        """
        Gtk.ListBoxRow.__init__(self)
        if title == "Separator":
            self.__label = None
            separator = Gtk.Separator.new(Gtk.Orientation.HORIZONTAL)
            separator.show()
            self.add(separator)
            self.set_sensitive(False)
        else:
            self.get_style_context().add_class("big-padding")
            self.__label = Gtk.Label.new(title)
            self.__label.show()
            self.__label.set_property("halign", Gtk.Align.START)
            self.add(self.__label)

    @property
    def title(self):
        """
            Get row title
            @return str
        """
        return self.__label.get_text() if self.__label is not None else None


class ComboBox(Gtk.MenuButton):
    """
        Implement one combobox to prevent this GTK bug:
        https://gitlab.gnome.org/World/lollypop/issues/2253
    """
    # Same signal than MenuWidget
    __gsignals__ = {
        "changed": (GObject.SignalFlags.RUN_FIRST, None, ()),
    }

    def __init__(self):
        """
            Init widget
        """
        Gtk.MenuButton.__init__(self)
        grid = Gtk.Grid.new()
        grid.show()
        self.__label = Gtk.Label.new()
        self.__label.show()
        image = Gtk.Image.new_from_icon_name("pan-down-symbolic",
                                             Gtk.IconSize.BUTTON)
        image.show()
        grid.add(self.__label)
        grid.add(image)
        self.set_image(grid)
        self.__popover = Gtk.Popover.new()
        self.__popover.set_relative_to(self)
        height = max(250, App().window.get_allocated_height() / 2)
        self.__popover.set_size_request(250, height)
        self.__scrolled = Gtk.ScrolledWindow.new()
        self.__scrolled.show()
        self.__scrolled.set_policy(Gtk.PolicyType.NEVER,
                                   Gtk.PolicyType.AUTOMATIC)
        self.__listbox = Gtk.ListBox.new()
        self.__listbox.show()
        self.__listbox.connect("row-activated", self.__on_row_activated)
        self.__scrolled.add(self.__listbox)
        self.__popover.add(self.__scrolled)
        self.set_popover(self.__popover)
        size_group = Gtk.SizeGroup.new(Gtk.SizeGroupMode.HORIZONTAL)
        size_group.add_widget(self.__label)
        size_group.add_widget(self.__popover)

    def append(self, text):
        """
            Appends text to the list of strings stored in self
            @param text as str
        """
        row = ComboRow(text)
        row.show()
        self.__listbox.add(row)
        self.__label.set_text(text)

    def set_label(self, text):
        """
            Set button label
            @param text as str
        """
        self.__label.set_text(text)

    def get_active_id(self):
        """
            Get active id
            @return str
        """
        return self.__listbox.get_selected_row().title

    def set_active_id(self, text):
        """
            Mark item_id as active
            @parma text as str
        """
        for row in self.__listbox.get_children():
            if row.title == text:
                self.__listbox.select_row(row)
                self.__label.set_text(text)
                row.hide()
                break
        GLib.idle_add(self.__hide_row, row)

#######################
# PRIVATE             #
#######################
    def __hide_row(self, row):
        """
            Hide row, show others
            @param row as ComboRow
        """
        for _row in self.__listbox.get_children():
            if row != _row:
                _row.show()
        row.hide()
        self.__scrolled.get_vadjustment().set_value(0)

    def __sort_listbox(self, rowa, rowb):
        """
            Sort listbox
            @param rowa as ComboRow
            @param rowb as ComboRow
        """
        return rowa.index > rowb.index

    def __on_row_activated(self, listbox, row):
        """
            Close popover and change label
            @param listbox as Gtk.ListBox
            @param row as ComboRow
        """
        self.__label.set_text(row.title)
        self.__popover.popdown()
        emit_signal(self, "changed")
        # Delay to let popover popdown
        GLib.timeout_add(500, self.__hide_row, row)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, GObject, GLib, Handy

from lollypop.helper_art import ArtBehaviour
from lollypop.define import App, ArtSize, MARGIN_SMALL, MARGIN
from lollypop.widgets_player_progress import ProgressPlayerWidget
from lollypop.widgets_player_buttons import ButtonsPlayerWidget
from lollypop.widgets_player_artwork import ArtworkPlayerWidget
from lollypop.widgets_player_label import LabelPlayerWidget
from lollypop.helper_size_allocation import SizeAllocationHelper
from lollypop.helper_signals import SignalsHelper, signals
from lollypop.utils import emit_signal


class MiniPlayer(Handy.WindowHandle, SizeAllocationHelper, SignalsHelper):
    """
        Mini player shown in folded window
    """
    __gsignals__ = {
        "revealed": (GObject.SignalFlags.RUN_FIRST, None, (bool,)),
    }

    @signals
    def __init__(self):
        """
            Init mini player
        """
        Handy.WindowHandle.__init__(self)
        SizeAllocationHelper.__init__(self)
        self.__previous_artwork_id = None
        self.__per_track_cover = App().settings.get_value(
            "allow-per-track-cover")
        self.get_style_context().add_class("black")
        self.__box = Gtk.Box.new(Gtk.Orientation.VERTICAL, 0)
        self.__box.show()
        self.__box.get_style_context().add_class("padding")
        self.__revealer = Gtk.Revealer.new()
        self.__revealer.show()
        self.__revealer_box = Gtk.Box.new(Gtk.Orientation.VERTICAL,
                                          MARGIN_SMALL)
        self.__revealer_box.show()
        bottom_box = Gtk.Box.new(Gtk.Orientation.HORIZONTAL, MARGIN_SMALL)
        bottom_box.show()
        self.__progress_widget = ProgressPlayerWidget()
        self.__progress_widget.show()
        self.__progress_widget.set_vexpand(True)
        self.__progress_widget.set_margin_start(MARGIN)
        self.__progress_widget.set_margin_end(MARGIN)
        buttons_widget = ButtonsPlayerWidget(["banner-button-big"])
        buttons_widget.show()
        buttons_widget.update()
        buttons_widget.set_property("valign", Gtk.Align.CENTER)
        buttons_widget.set_margin_end(MARGIN_SMALL)
        self.__artwork_widget = ArtworkPlayerWidget(ArtBehaviour.CROP_SQUARE)
        self.__artwork_widget.show()
        self.__artwork_widget.set_vexpand(True)
        self.__artwork_widget.set_property("margin", MARGIN_SMALL)
        self.__artwork_widget.get_style_context().add_class(
                "transparent-cover-frame")
        self.__label_box = Gtk.Box(Gtk.Orientation.HORIZONTAL, MARGIN)
        self.__label_box.show()
        self.__label_box.set_property("margin", MARGIN_SMALL)
        self.__label_widget = LabelPlayerWidget(False, 9)
        self.__label_widget.show()
        self.__label_widget.update()
        self.__label_widget.set_margin_start(MARGIN_SMALL)
        self.__label_widget.set_margin_end(MARGIN_SMALL)
        self.__artwork_button = Gtk.Image.new()
        self.__artwork_button.show()
        self.__label_box.pack_start(self.__artwork_button, False, False, 0)
        self.__label_box.pack_start(self.__label_widget, False, False, 0)
        button = Gtk.Button.new()
        button.show()
        button.get_style_context().add_class("banner-button-big")
        button.set_property("halign", Gtk.Align.START)
        button.set_property("valign", Gtk.Align.CENTER)
        button.connect("clicked", self.__on_button_clicked)
        button.set_image(self.__label_box)
        button.set_margin_start(MARGIN_SMALL)
        self.__background = Gtk.Image()
        self.__background.show()
        # Assemble UI
        self.__box.pack_start(self.__revealer, False, True, 0)
        self.__box.pack_start(bottom_box, False, False, 0)
        bottom_box.pack_start(button, True, True, 0)
        bottom_box.pack_end(buttons_widget, False, False, 0)
        self.__revealer.add(self.__revealer_box)
        self.__revealer_box.pack_start(self.__artwork_widget, False, True, 0)
        self.__revealer_box.pack_end(self.__progress_widget, False, True, 0)
        overlay = Gtk.Overlay.new()
        overlay.show()
        overlay.add(self.__background)
        overlay.add_overlay(self.__box)
        self.add(overlay)
        return [
            (App().player, "current-changed", "_on_current_changed")
        ]

    def reveal(self, reveal):
        """
            Reveal miniplayer
            @param reveal as bool
        """
        if reveal:
            self.__revealer.set_reveal_child(True)
            emit_signal(self, "revealed", True)
            self.__progress_widget.update()
            size = min(ArtSize.FULLSCREEN, self.width // 2)
            self.__artwork_widget.set_art_size(size, size)
            self.__artwork_widget.update(True)
            if App().lookup_action("miniplayer").get_state():
                self.__artwork_button.set_from_icon_name(
                    "view-fullscreen-symbolic", Gtk.IconSize.BUTTON)
            else:
                self.__artwork_button.set_from_icon_name(
                    "pan-down-symbolic", Gtk.IconSize.BUTTON)
        else:
            self.__revealer.set_reveal_child(False)
            emit_signal(self, "revealed", False)
            self.update_artwork()
        self.__set_widgets_position()

    def update_artwork(self):
        """
            Update artwork
        """
        if not self.revealed:
            App().art_helper.set_album_artwork(
                    App().player.current_track.album,
                    ArtSize.SMALL,
                    ArtSize.SMALL,
                    self.__artwork_button.get_scale_factor(),
                    ArtBehaviour.CACHE |
                    ArtBehaviour.CROP_SQUARE,
                    self.__on_button_artwork)

    def do_get_preferred_width(self):
        """
            Force preferred width
        """
        (min, nat) = Gtk.Bin.do_get_preferred_width(self)
        # Allow resizing
        return (0, nat)

    def do_get_preferred_height(self):
        """
            Force preferred height
        """
        (min, nat) = self.__box.get_preferred_height()
        return (min, min)

    @property
    def revealed(self):
        """
            True if mini player revealed
            @return bool
        """
        return self.__revealer.get_reveal_child()

#######################
# PROTECTED           #
#######################
    def _on_current_changed(self, player):
        """
            Update artwork and labels
            @param player as Player
        """
        if player.current_track.id is None:
            self.__on_artwork(None)
            self.__previous_artwork_id = None
            self.reveal(False)
            self.hide()
            size = App().settings.get_value("window-size")
            maximized = App().settings.get_value("window-maximized")
            App().window.resize(size[0], size[1])
            if maximized:
                GLib.idle_add(App().window.maximize)
            return

        same_artwork = self.__previous_artwork_id ==\
            player.current_track.album.id and not self.__per_track_cover
        if same_artwork:
            return
        self.__previous_artwork_id = App().player.current_track.album.id
        allocation = self.get_allocation()
        self.__artwork_widget.set_artwork(
                allocation.width,
                allocation.height,
                self.__on_artwork,
                ArtBehaviour.BLUR_HARD | ArtBehaviour.DARKER)
        self.update_artwork()

    def _handle_width_allocate(self, allocation):
        """
            Handle artwork sizing
            @param allocation as Gtk.Allocation
        """
        if SizeAllocationHelper._handle_width_allocate(self, allocation):
            # We use parent height because we may be collapsed
            parent = self.get_parent()
            if parent is None:
                height = self.height
            else:
                height = parent.get_allocated_height()
            if self.__revealer.get_reveal_child():
                size = min(ArtSize.FULLSCREEN, self.width // 2)
                self.__artwork_widget.set_art_size(size, size)
                self.__artwork_widget.update(True)
            self.__artwork_widget.set_artwork(
                allocation.width + 100, height + 100,
                self.__on_artwork,
                ArtBehaviour.BLUR_HARD | ArtBehaviour.DARKER)

    def _handle_height_allocate(self, allocation):
        """
            Handle artwork sizing
            @param allocation as Gtk.Allocation
        """
        if SizeAllocationHelper._handle_height_allocate(self, allocation):
            # We use parent height because we may be collapsed
            parent = self.get_parent()
            if parent is None:
                height = allocation.height
            else:
                height = parent.get_allocated_height()
            self.__artwork_widget.set_artwork(
                allocation.width + 100, height + 100,
                self.__on_artwork,
                ArtBehaviour.BLUR_HARD | ArtBehaviour.DARKER)

#######################
# PRIVATE             #
#######################
    def __set_widgets_position(self):
        """
            Add label widget to wanted UI part
        """
        parent = self.__label_widget.get_parent()
        if parent is not None:
            parent.remove(self.__label_widget)
        if self.revealed:
            self.__revealer_box.pack_start(self.__label_widget, False, True, 0)
        else:
            self.__label_box.pack_start(self.__label_widget, False, False, 0)

    def __on_button_clicked(self, *ignore):
        """
            Set revealer on/off
        """
        if App().lookup_action("miniplayer").get_state():
            App().lookup_action("miniplayer").change_state(
                GLib.Variant("b", False))
        else:
            self.reveal(not self.revealed)

    def __on_button_artwork(self, surface):
        """
           Set artwork on button
           @param surface as cairo.Surface
        """
        self.__artwork_button.set_from_surface(surface)
        del surface

    def __on_artwork(self, surface):
        """
            Set artwork
            @param surface as str
        """
        self.__background.set_from_surface(surface)
        if surface is None:
            self.__background.get_style_context().add_class("black")
        else:
            self.__background.get_style_context().remove_class("black")
            del surface



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import GLib, Gtk

from lollypop.utils import emit_signal
from lollypop.objects_track import Track
from lollypop.widgets_row_track import TrackRow
from lollypop.view_albums_list import AlbumsListView
from lollypop.view_tracks_queue import QueueTracksView
from lollypop.logger import Logger
from lollypop.define import App, ViewType, Size, MARGIN
from lollypop.helper_signals import SignalsHelper, signals_map
from lollypop.widgets_banner_current_albums import CurrentAlbumsBannerWidget


class CurrentAlbumsView(AlbumsListView, SignalsHelper):
    """
        Popover showing Albums View
    """

    @signals_map
    def __init__(self, view_type):
        """
            Init view
            @param view_type as ViewType
        """
        AlbumsListView.__init__(self, [], [],
                                view_type |
                                ViewType.SCROLLED |
                                ViewType.OVERLAY |
                                ViewType.PLAYBACK)
        self.box.set_width(Size.MEDIUM)
        if view_type & ViewType.DND:
            self.dnd_helper.connect("dnd-finished", self.__on_dnd_finished)
        self.__banner = CurrentAlbumsBannerWidget(self)
        self.__banner.show()
        # Queue
        self.__queue_widget = QueueTracksView()
        self.__queue_widget.show()
        self.__queue_widget.populate()
        grid = Gtk.Grid()
        grid.set_orientation(Gtk.Orientation.VERTICAL)
        grid.show()
        grid.add(self.__queue_widget)
        grid.add(self.box)
        self.add_widget(grid, self.__banner)
        self.allow_duplicate("_on_playback_added")
        self.allow_duplicate("_on_playback_updated")
        self.allow_duplicate("_on_playback_removed")
        return [
            (App().player, "playback-added", "_on_playback_added"),
            (App().player, "playback-setted", "_on_playback_setted"),
            (App().player, "playback-updated", "_on_playback_updated"),
            (App().player, "playback-removed", "_on_playback_removed")
        ]

    def populate(self):
        """
            Populate view
        """
        albums = App().player.albums
        if albums:
            if len(albums) == 1:
                self.add_reveal_albums(albums)
            AlbumsListView.populate(self, albums)
            self.show_placeholder(False)
        else:
            self.show_placeholder(True)

    def clear(self):
        """
            Clear the view
        """
        if not App().player.radio_cancellable.is_cancelled():
            App().player.radio_cancellable.cancel()
            GLib.timeout_add(500, self.clear)
        else:
            AlbumsListView.clear(self)
            App().player.clear_albums()

    @property
    def filtered(self):
        """
            Get filtered children
            @return [Gtk.Widget]
        """
        filtered = []
        for child in self.children:
            filtered.append(child)
            for subchild in child.children:
                filtered.append(subchild)
        return filtered

    @property
    def scroll_shift(self):
        """
            Get scroll shift for y axes
            @return int
        """
        return self.__banner.height + MARGIN

    @property
    def args(self):
        return None

#######################
# PROTECTED           #
#######################
    def _on_playback_added(self, player, album):
        """
            Add album
            @param player as Player
            @param album as Album
        """
        self.add_value(album)
        self.show_placeholder(False)

    def _on_playback_updated(self, player, album):
        """
            Reset album
            @param player as Player
            @param album as Album
        """
        for child in self.children:
            if child.album == album:
                child.reset()
                break

    def _on_playback_setted(self, player, albums):
        """
            Add album
            @param player as Player
            @param albums as [Album]
        """
        if albums:
            self.stop()
            AlbumsListView.clear(self)
            AlbumsListView.populate(self, albums)
            self.show_placeholder(False)
        else:
            self.stop()
            AlbumsListView.clear(self)
            self.show_placeholder(True)

    def _on_playback_removed(self, player, album):
        """
            Remove row
            @param player as Player
            @param album as Album
        """
        for child in self.children:
            if child.album == album:
                self._box.remove(child)
                break
        if not self.children:
            self.show_placeholder(True)

    def _on_row_activated(self, row, track):
        """
            Start playback
            @param row as AlbumRow
            @param track as Track
        """
        App().player.load(track)

    def _on_album_removed(self, row):
        """
            Remove album from playback
            @param row as AlbumRow
        """
        try:
            if row.album.id in App().player.album_ids:
                if App().player.current_track in row.album.tracks:
                    App().player.skip_album()
                App().player.remove_album(row.album)
            else:
                App().player.add_album(row.album)
        except Exception as e:
            Logger.error("CurrentAlbumsView::_on_album_removed()", e)

    def _on_track_removed(self, row, track):
        """
            Remove track from playback
            @param row as AlbumRow
            @param track as Track
        """
        if App().player.current_track == track:
            if App().player.next_track.id != track.id:
                App().player.next()
        row.album.remove_track(track)
        App().player.update_next_prev()
        emit_signal(App().player, "playback-updated", row.album)
        if not row.children:
            row.destroy()

#######################
# PRIVATE             #
#######################
    def __add_queue(self):
        """
            Add player queue
        """
        for track_id in App().player.queue:
            row = TrackRow(Track(track_id), [], self.view_type)
            row.show()
            self.__queue_widget.add(row)

    def __on_dnd_finished(self, dnd_helper):
        """
            Save playlist if needed
            @param dnd_helper as DNDHelper
        """
        albums = []
        for child in self.children:
            albums.append(child.album)
        App().player.set_albums(albums, False)



# Copyright (c) 2014-2017 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Soup

import json
from base64 import b64encode
from time import time, sleep
from hashlib import md5

from lollypop.logger import Logger
from lollypop.utils import get_network_available
from lollypop.define import SPOTIFY_CLIENT_ID, SPOTIFY_SECRET, App
from lollypop.define import LASTFM_API_KEY, LASTFM_API_SECRET
from lollypop.helper_passwords import PasswordsHelper


class TokenWebService:
    """
        Get token from web services
    """

    def __init__(self):
        """
            Init object
        """
        self.__token_expires = {"SPOTIFY": 0}
        self.__tokens = {"SPOTIFY": None,
                         "LASTFM": None,
                         "LIBREFM": None}
        self.__loading_token = {"SPOTIFY": False,
                                "LASTFM": False,
                                "LIBREFM": False}
        self.__passwords_helper = PasswordsHelper()

    def get_token(self, service, cancellable):
        """
            Get a token for name
            @param service as str
            @param cancellable as Gio.Cancellable
            @return str/None
        """
        if self.__tokens[service] == "" or not get_network_available():
            return None
        while self.__loading_token[service]:
            sleep(0.1)
        # Remove 60 seconds to be sure
        token_expired = service in self.__token_expires.keys() and\
            int(time()) + 60 > self.__token_expires[service]
        if token_expired or self.__tokens[service] is None:
            self.__loading_token[service] = True
            self.__get_token(service, cancellable)
            self.__loading_token[service] = False
        return self.__tokens[service]

    def get_lastfm_auth_token(self, service, cancellable, callback):
        """
            Get a new initial auth token
            @param service as str
            @param cancellable as Gio.Cancellable
            @param callback as function
            @thread safe
        """
        try:
            def on_data(uri, status, data):
                if status:
                    decode = json.loads(data.decode("utf-8"))
                    token = "validation:%s" % decode["token"]
                    callback(token, service)
            if service == "LASTFM":
                uri = "https://ws.audioscrobbler.com/2.0/"
            else:
                uri = "https://libre.fm/2.0/"
            uri += "?api_key=%s&format=json&method=auth.gettoken" %\
                LASTFM_API_KEY
            api_sig = "api_key%smethodauth.gettoken%s" % (
                LASTFM_API_KEY, LASTFM_API_SECRET)
            encoded = md5(api_sig.encode("utf-8")).hexdigest()
            uri += "&api_sig=%s" % encoded
            App().task_helper.load_uri_content(uri, cancellable, on_data)
        except Exception as e:
            Logger.error(
                "TokenWebService::get_lastfm_auth_token(): %s", e)

    def clear_token(self, service, clear_secret=False):
        """
            Clear token
            @param service as str
            @param clear_secret as bool
        """
        self.__tokens[service] = None
        if clear_secret:
            self.__passwords_helper.clear(service)

#######################
# PRIVATE             #
#######################
    def __load_token_for_service(self, token, service):
        """
            Load token for service
            @param token as str
            @param service as str
        """
        if token is None:
            self.__tokens[service] = ""
        elif token.startswith("validation:"):
            validation_token = token.replace("validation:", "")
            self.__get_lastfm_token(validation_token, service, None)
        else:
            self.__tokens[service] = token

    def __get_token(self, service, cancellable):
        """
            Get a token for name
            @param service as str
            @param cancellable as Gio.Cancellable
        """
        if service == "SPOTIFY":
            self.__get_spotify_token(cancellable)
        else:
            token = self.__passwords_helper.get_token(service)
            self.__load_token_for_service(token, service)

    def __get_spotify_token(self, cancellable):
        """
            Get a new auth token
            @param cancellable as Gio.Cancellable
        """
        try:
            uri = "https://accounts.spotify.com/api/token"
            credentials = "%s:%s" % (SPOTIFY_CLIENT_ID, SPOTIFY_SECRET)
            encoded = b64encode(credentials.encode("utf-8"))
            credentials = encoded.decode("utf-8")
            data = Soup.form_encode_hash({"grant_type": "client_credentials"})
            msg = Soup.Message.new_from_encoded_form("POST", uri, data)
            request_headers = msg.get_property("request-headers")
            request_headers.append("Authorization",
                                   "Basic %s" % credentials)
            data = App().task_helper.send_message_sync(msg, cancellable)
            if data is not None:
                decode = json.loads(data.decode("utf-8"))
                self.__token_expires["SPOTIFY"] = int(time()) +\
                    int(decode["expires_in"])
                self.__tokens["SPOTIFY"] = decode["access_token"]
        except Exception as e:
            Logger.error("TokenWebService::__get_spotify_token(): %s", e)
        self.__loading_token["SPOTIFY"] = False

    def __get_lastfm_token(self, token, service, cancellable):
        """
            Get token validated by user
            @param token as str
            @param service as str
            @param cancellable as Gio.Cancellable
        """
        try:
            if service == "LASTFM":
                uri = "https://ws.audioscrobbler.com/2.0/"
                uri += "?api_key=%s&format=json" % LASTFM_API_KEY
                uri += "&method=auth.getsession&token=%s" % token
                api_sig = "api_key%smethodauth.getsessiontoken%s%s" % (
                    LASTFM_API_KEY, token, LASTFM_API_SECRET)
                encoded = md5(api_sig.encode("utf-8")).hexdigest()
                uri += "&api_sig=%s" % encoded
            else:
                uri = "https://libre.fm/2.0/?format=json"
                uri += "&method=auth.getsession&token=%s" % token
            (status, data) = App().task_helper.load_uri_content_sync(
                uri, cancellable)
            if status:
                decode = json.loads(data.decode("utf-8"))
                self.__tokens[service] = decode["session"]["key"]
                self.__passwords_helper.clear(service,
                                              self.__passwords_helper.store,
                                              service,
                                              decode["session"]["name"],
                                              self.__tokens[service])
        except:
            self.__passwords_helper.clear(service)
            Logger.error(
                "TokenWebService::__get_lastfm_session(): %s", decode)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, GLib, GObject

from lollypop.utils import emit_signal, on_query_tooltip, get_network_available
from lollypop.define import App, ArtSize, ArtBehaviour, ViewType
from lollypop.widgets_banner import BannerWidget
from lollypop.helper_signals import SignalsHelper, signals_map


class InformationBannerWidget(BannerWidget, SignalsHelper):
    """
        Banner for Information
    """

    __gsignals__ = {
        "search": (GObject.SignalFlags.RUN_FIRST, None, (bool,)),
    }

    @signals_map
    def __init__(self, artist_id):
        """
            Init information banner
            @param artist_id as int
        """
        BannerWidget.__init__(self, ViewType.OVERLAY)
        self.__artist_id = artist_id
        builder = Gtk.Builder()
        builder.add_from_resource(
            "/org/gnome/Lollypop/InformationBannerWidget.ui")
        builder.connect_signals(self)
        self.__button = builder.get_object("button")
        self.__badge_artwork = builder.get_object("badge_artwork")
        self.__title_label = builder.get_object("artist")
        self.__title_label.connect("query-tooltip", on_query_tooltip)
        self.__title_label.set_property("has-tooltip", True)
        self.__widget = builder.get_object("widget")
        self.__title_label.set_markup(GLib.markup_escape_text(
            App().artists.get_name(artist_id)))
#        self.__title_label.get_style_context().add_class("text-x-large")
        self._overlay.add_overlay(self.__widget)
        self._overlay.set_overlay_pass_through(self.__widget, True)
        self.__set_internal_size()
        if not get_network_available("WIKIPEDIA"):
            self.__button.set_sensitive(False)
        return [
               (App().window.container.widget, "notify::folded",
                "_on_container_folded"),
        ]

    def update_for_width(self, width):
        """
            Update banner internals for width, call this before showing banner
            @param width as int
        """
        BannerWidget.update_for_width(self, width)
        self.__set_artwork()

    @property
    def button(self):
        """
            Get toggle button
            @return Gtk.ToggleButton
        """
        return self.__button

#######################
# PROTECTED           #
#######################
    def _handle_width_allocate(self, allocation):
        """
            Update artwork
            @param allocation as Gtk.Allocation
        """
        if BannerWidget._handle_width_allocate(self, allocation):
            self.__set_artwork()

    def _on_container_folded(self, leaflet, folded):
        """
            Handle libhandy folded status
            @param leaflet as Handy.Leaflet
            @param folded as Gparam
        """
        self.__set_internal_size()

    def _on_button_toggled(self, button):
        """
            Emit search signal
            @param button as Gtk.ToggleButton
        """
        if get_network_available("WIKIPEDIA"):
            emit_signal(self, "search", button.get_active())

#######################
# PRIVATE             #
#######################
    def __set_artwork(self):
        """
            Set artwork
        """
        if App().settings.get_value("artist-artwork") and App().animations:
            artist = App().artists.get_name(self.__artist_id)
            App().art_helper.set_artist_artwork(
                                        artist,
                                        # +100 to prevent resize lag
                                        self.width + 100,
                                        self.height,
                                        self.get_scale_factor(),
                                        ArtBehaviour.BLUR_HARD |
                                        ArtBehaviour.DARKER,
                                        self._on_artwork)
        else:
            self._artwork.get_style_context().add_class("default-banner")
        if self.width < ArtSize.BANNER * 3:
            if self.__widget.get_child_at(1, 0) == self.__title_label:
                self.__widget.remove(self.__title_label)
                self.__widget.attach(self.__title_label, 2, 2, 3, 1)
        elif self.__widget.get_child_at(2, 2) == self.__title_label:
            self.__widget.remove(self.__title_label)
            self.__widget.attach(self.__title_label, 1, 0, 1, 3)

    def __set_badge_artwork(self, art_size):
        """
            Set artist artwork on badge
            @param art_size as int
        """
        if App().settings.get_value("artist-artwork"):
            artist = App().artists.get_name(self.__artist_id)
            App().art_helper.set_artist_artwork(
                                        artist,
                                        art_size,
                                        art_size,
                                        self.get_scale_factor(),
                                        ArtBehaviour.ROUNDED |
                                        ArtBehaviour.CROP_SQUARE |
                                        ArtBehaviour.CACHE,
                                        self.__on_badge_artist_artwork,
                                        art_size)
        else:
            self.__badge_artwork.hide()

    def __set_internal_size(self):
        """
            Set content size based on current width
        """
        # Text size
        title_context = self.__title_label.get_style_context()
        for c in title_context.list_classes():
            title_context.remove_class(c)
        if App().window.folded:
            art_size = ArtSize.MEDIUM
            cls = "text-large"
        else:
            art_size = ArtSize.BANNER
            cls = "text-x-large"
        self.__title_label.get_style_context().add_class(cls)
        self.__set_badge_artwork(art_size)

    def __on_badge_artist_artwork(self, surface, art_size):
        """
            Set artist artwork on badge
            @param surface as cairo.Surface
            @param art_size as int
        """
        if surface is None:
            self.__badge_artwork.get_style_context().add_class("circle-icon")
            self.__badge_artwork.set_size_request(art_size, art_size)
            self.__badge_artwork.set_from_icon_name(
                                              "avatar-default-symbolic",
                                              Gtk.IconSize.DIALOG)
        else:
            self.__badge_artwork.get_style_context().remove_class(
                "circle-icon")
            self.__badge_artwork.set_from_surface(surface)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, GObject, Gdk, GdkPixbuf, GLib, Pango

from lollypop.objects_album import Album
from lollypop.define import App, ArtSize, ArtBehaviour, MARGIN
from lollypop.utils import get_round_surface, emit_signal
from lollypop.menu_header import HeaderType
from lollypop.helper_signals import SignalsHelper, signals_map


class MenuBuilder(Gtk.Stack, SignalsHelper):
    """
        Advanced menu model constructor
        Does not support submenus
    """

    __gsignals__ = {
        "hidden": (GObject.SignalFlags.RUN_FIRST, None, (bool,)),
    }

    @signals_map
    def __init__(self, menu, scrolled=False):
        """
            Init menu
            @param menu as Gio.Menu
            @param scrolled as bool
        """
        Gtk.Stack.__init__(self)
        self.__built = False
        self.__grids = {}
        self.__menu_queue = []
        self.__submenu_queue = []
        self.__widgets_queue = []
        self.__add_menu(menu, "main", False, scrolled)
        return [
            (App().window.container.widget, "notify::folded",
             "_on_container_folded")
        ]

    def add_widget(self, widget, position=-1):
        """
            Append widget to menu
            @param widget as Gtk.Widget
            @param position as negative int
        """
        self.__widgets_queue.append((widget, position))
        if self.__built:
            self.__add_widgets()

#######################
# PROTECTED           #
#######################
    def _on_container_folded(self, leaflet, folded):
        """
            Destroy self if adaptive off
            @param leaflet as Handy.Leaflet
            @param folded as Gparam
        """
        if not App().window.folded:
            self.destroy()

#######################
# PRIVATE             #
#######################
    def __add_widgets(self):
        """
            Add pending widget to menu
        """
        while self.__widgets_queue:
            (widget, position) = self.__widgets_queue.pop(0)
            if widget.submenu_name is not None:
                self.__add_menu_container(widget.submenu_name, True, True)
                self.__grids[widget.submenu_name].add(widget)
                button = Gtk.ModelButton.new()
                button.set_label(widget.submenu_name)
                button.get_child().set_halign(Gtk.Align.START)
                button.set_property("menu-name", widget.submenu_name)
                button.show()
                parent = self.__grids["main"]
                if widget.section is not None:
                    self.__add_section(widget.section, "main")
                widget = button
            else:
                parent = self.get_child_by_name("main")
                if isinstance(parent, Gtk.ScrolledWindow):
                    # scrolled -> viewport -> grid
                    parent = parent.get_child().get_child()
            if position < -1:
                position = len(parent) + position + 1
                parent.insert_row(position)
                parent.attach(widget, 0, position, 1, 1)
            else:
                parent.add(widget)

    def __add_menu_container(self, menu_name, submenu, scrolled, margin=10):
        """
            Add menu container
            @param menu_name as str
            @param submenu as bool
            @param scrolled as bool
            @param margin as int
        """
        grid = self.get_child_by_name(menu_name)
        if grid is None:
            grid = Gtk.Grid.new()
            grid.set_orientation(Gtk.Orientation.VERTICAL)
            grid.connect("map", self.__on_grid_map, menu_name)
            self.__grids[menu_name] = grid
            grid.set_property("margin", margin)
            grid.show()
            if scrolled:
                scrolled = Gtk.ScrolledWindow()
                scrolled.set_policy(Gtk.PolicyType.NEVER,
                                    Gtk.PolicyType.AUTOMATIC)
                scrolled.show()
                scrolled.add(grid)
                self.add_named(scrolled, menu_name)
            else:
                self.add_named(grid, menu_name)
            if submenu:
                button = Gtk.ModelButton.new()
                button.get_style_context().add_class("padding")
                button.set_property("menu-name", "main")
                button.set_property("inverted", True)
                button.set_label(menu_name)
                button.get_child().set_halign(Gtk.Align.START)
                button.show()
                grid.add(button)

    def __add_menu(self, menu, menu_name, submenu, scrolled):
        """
            Create container and add menu
            @param menu as Gio.Menu
            @param menu_name as str
            @param submenu as bool
            @param scrolled as bool
        """
        n_items = menu.get_n_items()
        if n_items:
            self.__add_menu_container(menu_name, submenu, scrolled, 10)
            menu_range = list(range(0, n_items))
            if submenu:
                self.__submenu_queue.append((menu, menu_name, menu_range))
            else:
                self.__add_menu_items(menu, menu_name, menu_range)
        else:
            self.__add_menu_container(menu_name, submenu, scrolled, 0)
            self.__built = True
            self.__add_widgets()

    def __add_menu_items(self, menu, menu_name, indexes):
        """
            Add menu items present in indexes
            @param menu as Gio.Menu
            @param menu_name as str
            @param indexes as [int]
        """
        if indexes:
            i = indexes.pop(0)
            header = menu.get_item_attribute_value(i, "header")
            action = menu.get_item_attribute_value(i, "action")
            label = menu.get_item_attribute_value(i, "label")
            tooltip = menu.get_item_attribute_value(i, "tooltip")
            close = menu.get_item_attribute_value(i, "close") is not None
            if header is not None:
                header_type = header[0]
                header_label = header[1]
                if header_type == HeaderType.ALBUM:
                    album_id = header[2]
                    self.__add_album_header(header_label,
                                            album_id,
                                            menu_name)
                elif header_type == HeaderType.ARTIST:
                    artist_id = header[2]
                    self.__add_artist_header(header_label, artist_id,
                                             menu_name)
                elif header_type == HeaderType.ROUNDED:
                    artwork_name = header[2]
                    self.__add_rounded_header(header_label, artwork_name,
                                              menu_name)
                else:
                    icon_name = header[2]
                    self.__add_header(header_label, icon_name, menu_name)
                GLib.idle_add(self.__add_menu_items, menu, menu_name, indexes)
            elif action is None:
                link = menu.get_item_link(i, "section")
                submenu = menu.get_item_link(i, "submenu")
                if link is not None:
                    self.__menu_queue.append((menu, menu_name, indexes))
                    self.__add_section(label, menu_name)
                    self.__add_menu(link, menu_name, False, False)
                elif submenu is not None:
                    self.__add_submenu(label, submenu, menu_name)
                    GLib.idle_add(self.__add_menu_items, menu,
                                  menu_name, indexes)
            else:
                target = menu.get_item_attribute_value(i, "target")
                self.__add_item(label, action, target,
                                tooltip, close, menu_name)
                GLib.idle_add(self.__add_menu_items, menu, menu_name, indexes)
        # Continue to populate queued menu
        elif self.__menu_queue:
            (menu, menu_name, indexes) = self.__menu_queue.pop(-1)
            GLib.idle_add(self.__add_menu_items, menu, menu_name, indexes)
        # Finish with submenus
        elif self.__submenu_queue:
            (menu, menu_name, indexes) = self.__submenu_queue.pop(-1)
            GLib.idle_add(self.__add_menu_items, menu, menu_name, indexes)
        else:
            self.__built = True
            self.__add_widgets()

    def __add_item(self, text, action, target, tooltip, close, menu_name):
        """
            Add a Menu item
            @param text as GLib.Variant
            @param action as Gio.Action
            @param target as GLib.Variant
            @parmam tooltip as GLib.Variant
            @param close as bool
            @param menu_name as str
        """
        button = Gtk.ModelButton.new()
        button.set_hexpand(True)
        button.set_action_name(action.get_string())
        button.set_label(text.get_string())
        button.get_child().set_halign(Gtk.Align.START)
        if close:
            button.connect("clicked",
                           lambda x: emit_signal(self, "hidden", True))
        if tooltip is not None:
            button.set_tooltip_markup(tooltip.get_string())
            button.set_has_tooltip(True)
        if target is not None:
            button.set_action_target_value(target)
        button.show()
        self.__grids[menu_name].add(button)

    def __add_section(self, text, menu_name):
        """
            Add section to menu
            @param text as as GLib.Variant
            @param menu_name as str
        """
        box = Gtk.Box.new(Gtk.Orientation.HORIZONTAL, 4)
        sep1 = Gtk.Separator.new(Gtk.Orientation.HORIZONTAL)
        sep1.set_hexpand(True)
        sep1.set_property("valign", Gtk.Align.CENTER)
        box.add(sep1)
        label = Gtk.Label.new(text.get_string())
        label.get_style_context().add_class("dim-label")
        if App().window.folded:
            label.get_style_context().add_class("padding")
        box.add(label)
        sep2 = Gtk.Separator.new(Gtk.Orientation.HORIZONTAL)
        sep2.set_property("valign", Gtk.Align.CENTER)
        sep2.set_hexpand(True)
        box.add(sep2)
        box.show_all()
        self.__grids[menu_name].add(box)

    def __add_submenu(self, text, menu, menu_name):
        """
            Add submenu
            @param text as GLib.Variant
            @param menu as Gio.Menu
            @param menu_name as str
        """
        submenu_name = text.get_string()
        self.__add_menu(menu, submenu_name, True, True)
        button = Gtk.ModelButton.new()
        button.set_hexpand(True)
        button.set_property("menu-name", submenu_name)
        button.set_label(text.get_string())
        button.get_child().set_halign(Gtk.Align.START)
        button.show()
        self.__grids[menu_name].add(button)

    def __add_header(self, text, icon_name, menu_name):
        """
            Add an header for albums to close menu
            @param text as GLib.Variant
            @param icon_name as GLib.Variant
            @param menu_name as str
        """
        button = Gtk.ModelButton.new()
        button.set_hexpand(True)
        button.connect("clicked", lambda x: emit_signal(self, "hidden", True))
        button.show()
        label = Gtk.Label.new()
        label.set_markup(text)
        label.set_ellipsize(Pango.EllipsizeMode.END)
        label.show()
        artwork = Gtk.Image.new_from_icon_name(icon_name,
                                               Gtk.IconSize.INVALID)
        artwork.set_pixel_size(ArtSize.SMALL)
        artwork.show()
        close_image = Gtk.Image.new_from_icon_name("pan-up-symbolic",
                                                   Gtk.IconSize.BUTTON)
        close_image.show()
        grid = Gtk.Grid()
        grid.set_column_spacing(MARGIN)
        grid.add(artwork)
        grid.add(label)
        grid.add(close_image)
        button.set_image(grid)
        button.get_style_context().add_class("padding")
        self.__grids[menu_name].add(button)

    def __add_album_header(self, text, album_id, menu_name):
        """
            Add an header for album to close menu
            @param text as str
            @param album_id as int
            @param menu_name as str
        """
        button = Gtk.ModelButton.new()
        button.set_hexpand(True)
        button.connect("clicked", lambda x: emit_signal(self, "hidden", True))
        button.show()
        label = Gtk.Label.new()
        label.set_markup(text)
        label.set_ellipsize(Pango.EllipsizeMode.END)
        label.show()
        artwork = Gtk.Image.new()
        close_image = Gtk.Image.new_from_icon_name("pan-up-symbolic",
                                                   Gtk.IconSize.BUTTON)
        close_image.show()
        grid = Gtk.Grid()
        grid.set_halign(Gtk.Align.START)
        grid.set_column_spacing(MARGIN)
        grid.add(artwork)
        grid.add(label)
        grid.add(close_image)
        button.set_image(grid)
        button.get_style_context().add_class("padding")
        App().art_helper.set_album_artwork(
                Album(album_id),
                ArtSize.SMALL,
                ArtSize.SMALL,
                artwork.get_scale_factor(),
                ArtBehaviour.CACHE | ArtBehaviour.CROP_SQUARE,
                self.__on_artwork,
                artwork)
        self.__grids[menu_name].add(button)

    def __add_artist_header(self, text, artist_id, menu_name):
        """
            Add an header for artist to close menu
            @param text as str
            @param artist_id as int
            @param menu_name as str
        """
        button = Gtk.ModelButton.new()
        button.set_hexpand(True)
        button.connect("clicked", lambda x: emit_signal(self, "hidden", True))
        button.show()
        label = Gtk.Label.new()
        label.set_markup(text)
        label.set_ellipsize(Pango.EllipsizeMode.END)
        label.show()
        artwork = Gtk.Image.new()
        close_image = Gtk.Image.new_from_icon_name("pan-up-symbolic",
                                                   Gtk.IconSize.BUTTON)
        close_image.show()
        grid = Gtk.Grid()
        grid.set_column_spacing(MARGIN)
        grid.add(artwork)
        grid.add(label)
        grid.add(close_image)
        button.set_image(grid)
        button.get_style_context().add_class("padding")
        artist_name = App().artists.get_name(artist_id)
        App().art_helper.set_artist_artwork(
                artist_name,
                ArtSize.SMALL,
                ArtSize.SMALL,
                artwork.get_scale_factor(),
                ArtBehaviour.CACHE |
                ArtBehaviour.CROP_SQUARE |
                ArtBehaviour.ROUNDED,
                self.__on_artwork,
                artwork)
        self.__grids[menu_name].add(button)

    def __add_rounded_header(self, text, artwork_name, menu_name):
        """
            Add an header for rounded widgets to close menu
            @param text as str
            @param artwork_name as str
            @param menu_name as str
        """
        def on_load_from_cache(pixbuf, artwork):
            if pixbuf is not None:
                scale_factor = artwork.get_scale_factor()
                surface = Gdk.cairo_surface_create_from_pixbuf(
                    pixbuf.scale_simple(ArtSize.MEDIUM, ArtSize.MEDIUM,
                                        GdkPixbuf.InterpType.BILINEAR),
                    scale_factor, None)
                rounded = get_round_surface(surface, scale_factor,
                                            ArtSize.MEDIUM / 4)
                artwork.set_from_surface(rounded)
                artwork.show()
        button = Gtk.ModelButton.new()
        button.set_hexpand(True)
        button.connect("clicked", lambda x: emit_signal(self, "hidden", True))
        button.show()
        label = Gtk.Label.new()
        label.set_ellipsize(Pango.EllipsizeMode.END)
        label.set_markup(text)
        label.show()
        artwork = Gtk.Image.new()
        artwork.get_style_context().add_class("light-background")
        close_image = Gtk.Image.new_from_icon_name("pan-up-symbolic",
                                                   Gtk.IconSize.BUTTON)
        close_image.show()
        grid = Gtk.Grid()
        grid.set_column_spacing(MARGIN)
        grid.add(artwork)
        grid.add(label)
        grid.add(close_image)
        button.set_image(grid)
        button.get_style_context().add_class("padding")
        App().task_helper.run(
                App().art.get_from_cache,
                artwork_name,
                "ROUNDED",
                ArtSize.BANNER, ArtSize.BANNER,
                callback=(on_load_from_cache, artwork))
        self.__grids[menu_name].add(button)

    def __on_artwork(self, surface, artwork):
        """
            Set artwork
            @param surface as str
            @param artwork as Gtk.Image
        """
        if surface is None:
            artwork.set_from_icon_name("folder-music-symbolic",
                                       Gtk.IconSize.BUTTON)
        else:
            artwork.set_from_surface(surface)
        artwork.show()

    def __on_grid_map(self, widget, menu_name):
        """
            On map, set stack order
            @param widget as Gtk.Widget
            @param menu_name as str
        """
        if menu_name == "main":
            self.set_transition_type(Gtk.StackTransitionType.SLIDE_LEFT)
            self.set_size_request(-1, -1)
        else:
            self.set_size_request(300, 400)
            self.set_transition_type(Gtk.StackTransitionType.SLIDE_RIGHT)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gio, GLib

from lollypop.objects_album import Album
from lollypop.define import App


class HeaderType:
    DEFAULT = 1
    ARTIST = 2
    ALBUM = 3
    ROUNDED = 4


class MenuHeader(Gio.MenuItem):
    """
        A simple menu header with label and icon
    """

    def __init__(self, label, icon_name):
        """
            Init menu
            @param label as str
            @param icon_name as str
        """
        Gio.MenuItem.__init__(self)
        header_type = GLib.Variant("i", HeaderType.DEFAULT)
        vlabel = GLib.Variant("s", label)
        vicon_name = GLib.Variant("s", icon_name)
        header = [header_type, vlabel, vicon_name]
        self.set_attribute_value("header", GLib.Variant("av", header))


class ArtistMenuHeader(Gio.MenuItem):
    """
        A menu header item for an artist
    """

    def __init__(self, artist_id):
        """
            Init menu
            @param artist_id as int
        """
        Gio.MenuItem.__init__(self)
        header_type = GLib.Variant("i", HeaderType.ARTIST)
        name = App().artists.get_name(artist_id)
        label = "<span alpha='40000'>%s</span>" % GLib.markup_escape_text(name)
        vlabel = GLib.Variant("s", label)
        vartist_id = GLib.Variant("i", artist_id)
        header = [header_type, vlabel, vartist_id]
        self.set_attribute_value("header", GLib.Variant("av", header))


class RoundedMenuHeader(Gio.MenuItem):
    """
        A menu header item for a playlist
    """

    def __init__(self, name, artwork_name):
        """
            Init menu
            @param name as str
            @param artwork_name as str
        """
        Gio.MenuItem.__init__(self)
        header_type = GLib.Variant("i", HeaderType.ROUNDED)
        label = "<span alpha='40000'>%s</span>" % GLib.markup_escape_text(name)
        vlabel = GLib.Variant("s", label)
        vartwork_name = GLib.Variant("s", artwork_name)
        header = [header_type, vlabel, vartwork_name]
        self.set_attribute_value("header", GLib.Variant("av", header))


class AlbumMenuHeader(Gio.MenuItem):
    """
        A menu header item for Albums/Tracks
    """

    def __init__(self, object):
        """
            Init menu
            @param object as Album/Track
        """
        Gio.MenuItem.__init__(self)
        header_type = GLib.Variant("i", HeaderType.ALBUM)
        if isinstance(object, Album):
            label = "<span alpha='40000'>%s</span>" % GLib.markup_escape_text(
                object.name)
            album_id = object.id
        else:
            label = "<b>%s</b>\n<span alpha='40000'>%s</span>" % (
                GLib.markup_escape_text(", ".join(object.artists)),
                GLib.markup_escape_text(object.name))
            album_id = object.album.id
        vlabel = GLib.Variant("s", label)
        valbum_id = GLib.Variant("i", album_id)
        header = [header_type, vlabel, valbum_id]
        self.set_attribute_value("header", GLib.Variant("av", header))



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gdk, GLib, Gio
from gettext import gettext as _

from lollypop.define import App, ArtSize, Notifications


class NotificationManager:
    """
        Freedesktop notification support
    """

    def __init__(self):
        """
            Init notification object with lollypop infos
        """
        self.__notification_timeout_id = None
        App().player.connect("current-changed", self.__on_current_changed)
        self.__notification = Gio.Notification.new("")
        self.__action = Gio.Notification.new("")
        self.__action.add_button_with_target(
            _("Previous"),
            "app.shortcut",
            GLib.Variant("s", "prev"))
        self.__action.add_button_with_target(
            _("Next"),
            "app.shortcut",
            GLib.Variant("s", "next"))

    def send_track(self, track):
        """
            Send a message about track
            @param track as Track
        """
        if App().settings.get_enum("notifications") == Notifications.NONE:
            return
        state = App().window.get_window().get_state()
        if track.id is None or\
                state & Gdk.WindowState.FOCUSED or\
                App().is_fullscreen:
            return

        cover_path = App().album_art.get_cache_path(track.album,
                                                    ArtSize.BIG,
                                                    ArtSize.BIG)
        if cover_path is None:
            icon = Gio.Icon.new_for_string("org.gnome.Lollypop-symbolic")
        else:
            f = Gio.File.new_for_path(cover_path)
            icon = Gio.FileIcon.new(f)
        self.__action.set_icon(icon)
        self.__action.set_title(track.title)
        if track.album.name == "":
            self.__action.set_body(", ".join(track.artists))
        else:
            self.__action.set_body("%s - %s" % (", ".join(track.artists),
                                                track.album.name))
        App().send_notification("current-changed", self.__action)
        if self.__notification_timeout_id is not None:
            GLib.source_remove(self.__notification_timeout_id)
        self.__notification_timeout_id = GLib.timeout_add(
            5000, self.__withdraw_notification)

    def send(self, title, body=""):
        """
            Send message to user
            @param title as str
            @param body as str
        """
        self.__notification.set_title(title)
        if body:
            self.__notification.set_body(body)
        App().send_notification("send-message", self.__notification)

#######################
# PRIVATE             #
#######################
    def __withdraw_notification(self):
        """
            Remove notification
        """
        self.__notification_timeout_id = None
        App().withdraw_notification("current-changed")

    def __on_current_changed(self, player):
        """
            Send notification with track_id infos
            @param player Player
        """
        if App().settings.get_enum("notifications") != Notifications.ALL:
            return
        self.send_track(player.current_track)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gst, GstPbutils

from lollypop.logger import Logger


class Codecs:
    """
        Handle missing gstreamer codecs
    """

    def __init__(self):
        """
            Init installer
        """
        self._messages = []

    def install(self):
        """
            Install missing plugins
        """
        if not GstPbutils.install_plugins_supported():
            return
        try:
            context = GstPbutils.InstallPluginsContext.new()
            try:
                context.set_desktop_id("org.gnome.Lollypop.desktop")
            except:
                pass  # Not supported by Ubuntu VIVID
            details = []
            for message in self._messages:
                detail = \
                    GstPbutils.missing_plugin_message_get_installer_detail(
                        message)
                details.append(detail)
            GstPbutils.install_plugins_async(
                details,
                context,
                self.__null)
        except Exception as e:
            Logger.error("Codecs::__init__(): %s" % e)

    def append(self, message):
        """
            Append a message
            @param message as Gst.Message
        """
        self._messages.append(message)

    def is_missing_codec(self, message):
        """
            Check if codec is missing
            @return missing as bool
        """
        error, debug = message.parse_error()
        if error.matches(Gst.CoreError.quark(),
                         Gst.CoreError.MISSING_PLUGIN):
            return True
        return False

#######################
# PRIVATE             #
#######################
    def __null(self, arg):
        pass



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from locale import getlocale, strcoll
from importlib import import_module

# Ugly magic to dynamically adapt to the current locale...
try:
    # Try loading 'index_of' function of the current locale
    _module = import_module('lollypop.locales.%s' % getlocale()[0].lower())
    index_of = _module.index_of
except:
    # In case the locale doesn't need special care, fall back to a naive
    # implementation
    def index_of(string):
        """
            Get index of a string in a locale-aware manner.
            This is the fallback, which simply returns the first character.
            @param string as str
            @return str
        """
        if string:
            return string[0]
        else:
            return ""


class LocalizedCollation(object):
    """
        COLLATE LOCALIZED missing from default sqlite installation
        Android only
    """

    def __init__(self):
        pass

    def __call__(self, v1, v2):
        i1 = index_of(v1).upper()
        i2 = index_of(v2).upper()
        if strcoll(i1, i2) < 0:
            return -1
        elif strcoll(i1, i2) == 0:
            return strcoll(v1, v2)
        else:
            return 1



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, GLib

from gettext import gettext as _

from lollypop.define import ViewType, App, MARGIN_SMALL, Type
from lollypop.helper_adaptive import AdaptiveHelper
from lollypop.helper_signals import SignalsHelper, signals_map
from lollypop.helper_filtering import FilteringHelper


class Reveal:
    NONE = 0
    TRUE = 1
    FALSE = 2

REVEAL_DELTA=300

class View(Gtk.Grid, AdaptiveHelper, FilteringHelper, SignalsHelper):
    """
        Generic view
    """

    @signals_map
    def __init__(self, storage_type, view_type):
        """
            Init view
            @param storage_type as StorageType
            @param view_type as ViewType
        """
        Gtk.Grid.__init__(self)
        AdaptiveHelper.__init__(self)
        FilteringHelper.__init__(self)
        self.__storage_type = storage_type
        self.__view_type = view_type
        self.__scrolled_position = None
        self.__destroyed = False
        self.__banner = None
        self.__placeholder = None
        self.scrolled_value = 0
        self.set_orientation(Gtk.Orientation.VERTICAL)
        self.set_border_width(0)
        self.__new_ids = []
        self._empty_message = _("No items to show")
        self._empty_icon_name = "folder-music-symbolic"

        if view_type & ViewType.SCROLLED:
            self.__scrolled = Gtk.ScrolledWindow.new()
            self.__event_controller = Gtk.EventControllerMotion.new(
                self.__scrolled)
            self.__event_controller.set_propagation_phase(
                Gtk.PropagationPhase.TARGET)
            self.__event_controller.connect("leave", self._on_view_leave)
            self.__scrolled.get_vadjustment().connect("value-changed",
                                                      self._on_value_changed)
            self.__scrolled.show()
            self.__scrolled.set_property("expand", True)

        # Stack for placeholder
        self.__stack = Gtk.Stack.new()
        self.__stack.show()
        self.__stack.set_transition_type(Gtk.StackTransitionType.NONE)

        self.connect("destroy", self._on_destroy)
        self.connect("map", self._on_map)
        self.connect("unmap", self._on_unmap)
        self.connect("realize", self._on_realize)
        return [
            (App().window.container.widget, "notify::folded",
             "_on_container_folded"),
        ]

    def add_widget(self, widget, banner=None):
        """
            Add widget to view
            Add banner if ViewType.OVERLAY
            @param widget as Gtk.Widget
        """
        self.__stack.add_named(widget, "main")
        if self.view_type & ViewType.OVERLAY:
            self.__overlay = Gtk.Overlay.new()
            self.__overlay.show()
            if self.view_type & ViewType.SCROLLED:
                self.__overlay.add(self.scrolled)
                self.__scrolled.add(self.__stack)
            else:
                self.__overlay.add(self.__stack)
            if banner is not None:
                self.__overlay.add_overlay(banner)
                self.__banner = banner
                self.__banner.connect("scroll", self.__on_banner_scroll)
            self.add(self.__overlay)
        elif self.view_type & ViewType.SCROLLED:
            self.__scrolled.add(self.__stack)
            if banner is not None:
                self.__banner = banner
                self.add(self.__banner)
            self.add(self.scrolled)
        else:
            if banner is not None:
                self.__banner = banner
                self.add(self.__banner)
            self.add(self.__stack)
        if banner is not None:
            if self.view_type & ViewType.SCROLLED:
                self.scrolled.get_vscrollbar().set_margin_top(
                    self.__banner.height)
            banner.connect("height-changed", self.__on_banner_height_changed)

    def populate(self):
        pass

    def pause(self):
        pass

    def stop(self):
        pass

    def show_placeholder(self, show, new_text=None):
        """
            Show placeholder
            @param show as bool
        """
        if show:
            if self.__placeholder is not None:
                GLib.timeout_add(200, self.__placeholder.destroy)
            message = new_text\
                if new_text is not None\
                else self._empty_message
            from lollypop.widgets_placeholder import Placeholder
            self.__placeholder = Placeholder(message, self._empty_icon_name)
            self.__placeholder.show()
            self.__stack.add(self.__placeholder)
            self.__stack.set_visible_child(self.__placeholder)
        else:
            self.__stack.set_visible_child_name("main")

    def set_scrolled(self, scrolled):
        """
            Add an external scrolled window
            @param scrolled as Gtk.ScrolledWindow
        """
        self.__scrolled = scrolled
        self.__view_type |= ViewType.SCROLLED

    def set_populated_scrolled_position(self, position):
        """
            Set scrolled position on populated
            @param position as int
        """
        if self.view_type & ViewType.SCROLLED:
            self.__scrolled_position = position

    @property
    def scrolled(self):
        """
            Get scrolled widget
            @return Gtk.ScrolledWindow
        """
        if self.view_type & ViewType.SCROLLED:
            return self.__scrolled
        else:
            return Gtk.ScrolledWindow.new()

    @property
    def banner(self):
        """
            Get view banner
            @return BannerWidget
        """
        return self.__banner

    @property
    def children(self):
        """
            Get view children
            @return [Gtk.Widget]
        """
        return []

    @property
    def storage_type(self):
        """
            Get storage type
            @return StorageType
        """
        return self.__storage_type

    @property
    def view_type(self):
        """
            View type less sizing
            @return ViewType
        """
        return self.__view_type

    @property
    def position(self):
        """
            Get scrolled position
            @return float
        """
        if self.view_type & ViewType.SCROLLED:
            position = self.scrolled.get_vadjustment().get_value()
        else:
            position = 0
        return position

    @property
    def destroyed(self):
        """
            True if widget has been destroyed
            @return bool
        """
        return self.__destroyed

#######################
# PROTECTED           #
#######################
    def _on_view_leave(self, event_controller):
        pass

    def _on_container_folded(self, leaflet, folded):
        """
            Handle libhandy folded status
            @param leaflet as Handy.Leaflet
            @param folded as Gparam
        """
        if self.__placeholder is not None and self.__placeholder.is_visible():
            self.__placeholder.set_folded(App().window.folded)

    def _on_value_changed(self, adj):
        """
            Reveal banner
            @param adj as Gtk.Adjustment
        """
        if self.__banner is not None:
            reveal = self.__should_reveal_header(adj)
            if reveal == Reveal.TRUE:
                self.__banner.set_reveal_child(True)
                main_widget = self.__stack.get_child_by_name("main")
                if main_widget is not None:
                    main_widget.set_margin_top(self.__banner.height +
                                               MARGIN_SMALL)
            elif reveal == Reveal.FALSE:
                self.__banner.set_reveal_child(False)

    def _on_map(self, widget):
        """
            Set initial view state
            @param widget as GtK.Widget
        """
        # Set sidebar id
        if self.sidebar_id is None:
            ids = App().window.container.sidebar.selected_ids
            if ids:
                self.set_sidebar_id(ids[0])
                if self.sidebar_id == Type.GENRES_LIST:
                    self.selection_ids["left"] =\
                        App().window.container.left_list.selected_ids
                    self.selection_ids["right"] =\
                        App().window.container.right_list.selected_ids
                elif self.sidebar_id == Type.ARTISTS_LIST:
                    self.selection_ids["left"] =\
                        App().window.container.left_list.selected_ids

    def _on_unmap(self, widget):
        pass

    def _on_realize(self, widget):
        """
            Delayed adaptive mode
            Restore scroll position
            @param widget as Gtk.Widget
        """
        parent = widget.get_parent()
        if self.__banner is not None and parent is not None:
            width = parent.get_allocated_width()
            self.__banner.update_for_width(width)
            self.__on_banner_height_changed(self.__banner,
                                            self.__banner.height)
        # Wait for stack allocation to restore scrolled position
        if self.__scrolled_position is not None:
            self.__stack.connect("size-allocate",
                                 self.__on_stack_size_allocated)

    def _on_destroy(self, widget):
        """
            Clean up widget
            @param widget as Gtk.Widget
        """
        self.__destroyed = True
        self.__event_controller = None

#######################
# PRIVATE             #
#######################
    def __should_reveal_header(self, adj):
        """
            Check if we need to reveal header
            @param adj as Gtk.Adjustment
            @param delta as int
            @return int
        """
        value = adj.get_value()
        if (self.scrolled_value > value + REVEAL_DELTA):
            reveal = Reveal.TRUE
            self.scrolled_value = value
        elif (self.scrolled_value < value - REVEAL_DELTA):
            reveal = Reveal.FALSE
            self.scrolled_value = value
        else:
            reveal = Reveal.NONE

        return reveal

    def __on_banner_height_changed(self, banner, height):
        """
            Update scroll margin
            @param banner as BannerWidget
            @param height as int
        """
        if self.view_type & ViewType.OVERLAY:
            main_widget = self.__stack.get_child_by_name("main")
            main_widget.set_margin_top(height + MARGIN_SMALL)
            if self.view_type & ViewType.SCROLLED:
                self.scrolled.get_vscrollbar().set_margin_top(height)

    def __on_banner_scroll(self, banner, x, y):
        """
            Pass to scrolled
            @param banner as BannerWidget
            @param x as float
            @param y as float
        """
        if y > 0:
            y = 100
        else:
            y = -100
        adj = self.scrolled.get_vadjustment()
        new_value = adj.get_value() + y
        lower = adj.get_lower()
        upper = adj.get_upper() - adj.get_page_size()
        if new_value != lower and new_value != upper:
            adj.set_value(new_value)

    def __on_stack_size_allocated(self, stack, allocation):
        """
            Restore scrolled position
            @param stack as Gtk.Stack
            @param allocation as Gdk.Rectangle
        """
        if self.__scrolled_position is not None and\
                allocation.height > self.__scrolled_position:
            stack.disconnect_by_func(self.__on_stack_size_allocated)
            self.scrolled.get_vadjustment().set_value(
                self.__scrolled_position)
            self.__scrolled_position = None



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, Pango, GLib

from gettext import gettext as _

from lollypop.define import App, ArtSize, Type, ViewType
from lollypop.define import ArtBehaviour, MARGIN_MEDIUM, MARGIN, MARGIN_SMALL
from lollypop.widgets_rating import RatingWidget
from lollypop.widgets_loved import LovedWidget
from lollypop.widgets_label import LabelWidget
from lollypop.widgets_cover import EditCoverWidget, CoverWidget
from lollypop.widgets_banner import BannerWidget
from lollypop.utils import get_human_duration, on_query_tooltip
from lollypop.utils import popup_widget, emit_signal
from lollypop.helper_signals import SignalsHelper, signals_map


class AlbumBannerWidget(BannerWidget, SignalsHelper):
    """
        Banner for album
    """

    @signals_map
    def __init__(self, album, storage_type, view_type):
        """
            Init cover widget
            @param album
            @param storage_type as int
            @param view_type as ViewType
        """
        BannerWidget.__init__(self, view_type)
        self.__album = album
        self.__storage_type = storage_type
        self.__widget = None
        return [
                (App().window.container.widget, "notify::folded",
                 "_on_container_folded"),
                (App().album_art, "album-artwork-changed",
                 "_on_album_artwork_changed"),
                (App().player, "playback-added", "_on_playback_changed"),
                (App().player, "playback-updated", "_on_playback_changed"),
                (App().player, "playback-setted", "_on_playback_changed"),
                (App().player, "playback-removed", "_on_playback_changed")
        ]

    def populate(self):
        """
            Populate the view
        """
        if self.__widget is not None:
            return
        self.__widget = Gtk.Grid.new()
        self.__widget.set_margin_start(MARGIN)
        self.__widget.set_margin_end(MARGIN)
        self.__widget.set_margin_top(MARGIN_SMALL)
        self.__widget.set_margin_bottom(MARGIN_SMALL)
        self.__widget.set_row_spacing(MARGIN_SMALL)
        self.__widget.set_column_spacing(MARGIN_MEDIUM)
        self.__top_box = Gtk.Box.new(Gtk.Orientation.HORIZONTAL, 0)
        self.__middle_box = Gtk.Box.new(Gtk.Orientation.HORIZONTAL, 0)
        self.__bottom_box = Gtk.Box.new(Gtk.Orientation.HORIZONTAL, 0)
        self.__labels_box = Gtk.Box.new(Gtk.Orientation.VERTICAL, 0)
        self.__top_box.set_vexpand(True)
        self.__middle_box.set_halign(Gtk.Align.END)
        self.__middle_box.set_valign(Gtk.Align.CENTER)
        self.__middle_box.set_hexpand(True)
        self.__bottom_box.set_vexpand(True)
        self.__year_label = LabelWidget()
        self.__year_label.connect("clicked", self.__on_year_clicked)
        self.__year_label.set_justify(Gtk.Justification.RIGHT)
        self.__year_label.get_style_context().add_class("dim-label")
        self.__year_label.set_halign(Gtk.Align.END)
        self.__year_label.set_hexpand(True)
        self.__play_button = Gtk.Button.new_from_icon_name(
            "media-playback-start-symbolic", Gtk.IconSize.BUTTON)
        self.__add_button = Gtk.Button.new_from_icon_name(
            "list-add-symbolic", Gtk.IconSize.BUTTON)
        self.__menu_button = Gtk.Button.new_from_icon_name(
            "view-more-symbolic", Gtk.IconSize.BUTTON)
        self.__play_button.connect("clicked", self.__on_play_button_clicked)
        self.__add_button.connect("clicked", self.__on_add_button_clicked)
        self.__menu_button.connect("clicked", self.__on_menu_button_clicked)
        self.__title_label = Gtk.Label.new()
        self.__title_label.set_valign(Gtk.Align.END)
        self.__title_label.set_vexpand(True)
        self.__title_label.connect("query-tooltip", on_query_tooltip)
        self.__title_label.set_ellipsize(Pango.EllipsizeMode.END)
        self.__title_label.set_halign(Gtk.Align.START)
        self.__title_label.set_xalign(0)
        self.__artist_label = LabelWidget()
        self.__artist_label.show()
        self.__artist_label.connect("clicked", self.__on_artist_clicked)
        self.__artist_label.set_valign(Gtk.Align.START)
        self.__artist_label.set_halign(Gtk.Align.START)
        self.__artist_label.connect("query-tooltip", on_query_tooltip)
        self.__artist_label.set_ellipsize(Pango.EllipsizeMode.END)
        self.__artist_label.set_xalign(0)
        self.__duration_label = Gtk.Label.new()
        self.__duration_label.get_style_context().add_class("dim-label")
        self.__top_box.pack_end(self.__year_label, False, True, 0)
        self.__middle_box.add(self.__play_button)
        self.__middle_box.add(self.__add_button)
        self.__middle_box.add(self.__menu_button)
        self.__middle_box.get_style_context().add_class("linked")
        self.__bottom_box.pack_end(self.__duration_label, False, True, 0)
        self.__labels_box.add(self.__title_label)
        self.__labels_box.add(self.__artist_label)
        self.__widget.attach(self.__top_box, 2, 0, 1, 1)
        self.__set_album_year(True)
        self.__widget.attach(self.__middle_box, 2, 1, 1, 1)
        self.__widget.attach(self.__bottom_box, 1, 2, 2, 1)
        self.__widget.attach(self.__labels_box, 1, 0, 1, 2)
        self.__widget.show_all()
        if self.view_type & ViewType.OVERLAY:
            style = "banner-button"
        else:
            style = "menu-button"
        self.__play_button.get_style_context().add_class(style)
        self.__add_button.get_style_context().add_class(style)
        self.__menu_button.get_style_context().add_class(style)
        self.__title_label.set_label(self.__album.name)
        if self.view_type & ViewType.ALBUM:
            self.__cover_widget = EditCoverWidget(self.__album, self.view_type)
            self.__artist_label.get_style_context().add_class("dim-label")
            self.__artist_label.set_label(", ".join(self.__album.artists))
        else:
            self.__artist_label.hide()
            self.__cover_widget = CoverWidget(self.__album, self.view_type)
            self.__cover_widget.set_margin_top(MARGIN_MEDIUM)
            self.__cover_widget.set_margin_bottom(MARGIN_MEDIUM)
            self.__cover_widget.connect("populated",
                                        self.__on_cover_populated)
        self.__cover_widget.show()
        human_duration = get_human_duration(self.__album.duration)
        self.__duration_label.set_text(human_duration)
        self.__widget.attach(self.__cover_widget, 0, 0, 1, 3)
        self.__loved_widget = LovedWidget(self.__album, Gtk.IconSize.INVALID)
        self.__loved_widget.show()
        self.__bottom_box.pack_start(self.__loved_widget, 0, False, False)
        self.__rating_widget = RatingWidget(self.__album, Gtk.IconSize.INVALID)
        self.__rating_widget.show()
        self.__bottom_box.pack_start(self.__rating_widget, 0, True, True)
        if self.view_type & ViewType.OVERLAY:
            self._overlay.add_overlay(self.__widget)
            self._overlay.set_overlay_pass_through(self.__widget, True)
        else:
            self.add(self.__widget)
        self.__update_add_button()
        self.__set_internal_size()

    def update_for_width(self, width):
        """
            Update banner internals for width, call this before showing banner
            @param width as int
        """
        BannerWidget.update_for_width(self, width)
        self.__set_artwork()

    def set_selected(self, selected):
        """
            Mark widget as selected
            @param selected as bool
        """
        buttons = [self.__play_button, self.__add_button, self.__menu_button]
        for button in buttons:
            if selected:
                button.set_state_flags(Gtk.StateFlags.SELECTED, True)
            else:
                button.set_state_flags(Gtk.StateFlags.NORMAL, True)

    def update_duration(self):
        """
            Update duration label
        """
        human_duration = get_human_duration(self.__album.duration)
        self.__duration_label.set_text(human_duration)

#######################
# PROTECTED           #
#######################
    def _handle_width_allocate(self, allocation):
        """
            Update artwork
            @param allocation as Gtk.Allocation
        """
        if BannerWidget._handle_width_allocate(self, allocation):
            self.__set_artwork()

    def _on_container_folded(self, leaflet, folded):
        """
            Handle libhandy folded status
            @param leaflet as Handy.Leaflet
            @param folded as Gparam
        """
        self.__set_internal_size()

    def _on_album_artwork_changed(self, art, album_id):
        """
            Update cover for album_id
            @param art as Art
            @param album_id as int
        """
        if album_id == self.__album.id:
            self.__set_artwork()

    def _on_playback_changed(self, *ignore):
        """
            Update add button
        """
        self.__update_add_button()

#######################
# PRIVATE             #
#######################
    def __close_artwork_menu(self, action, variant):
        if App().window.folded:
            App().window.container.go_back()
        else:
            self.__artwork_popup.destroy()

    def __set_album_year(self, full):
        """
            Set album year
            @param full as int
        """
        if self.__album.year is not None:
            album_year = GLib.markup_escape_text(
                "%s" % self.__album.year)
            markup = None
            if full and self.__album.original_year is not None:
                original_year = GLib.markup_escape_text(
                    "%s" % self.__album.original_year)
                if original_year != album_year:
                    original_str = _("Released on %s") % original_year
                    markup = \
                        "%s\n<span size='x-small' alpha='40000'>%s</span>" %\
                        (album_year, original_str)
            if markup is None:
                markup = album_year
            self.__year_label.set_markup(markup)
            self.__year_label.show()

    def __set_artwork(self):
        """
            Set artwork on banner
        """
        if self._artwork is not None and\
                self.view_type & ViewType.ALBUM and\
                App().animations:
            App().art_helper.set_album_artwork(
                            self.__album,
                            # +100 to prevent resize lag
                            self.width + 100,
                            self.height,
                            self._artwork.get_scale_factor(),
                            ArtBehaviour.BLUR_HARD |
                            ArtBehaviour.DARKER,
                            self._on_artwork)
        if self.width < ArtSize.BANNER * 3:
            if self.__widget.get_child_at(2, 0) == self.__top_box:
                self.__widget.remove(self.__labels_box)
                self.__widget.remove(self.__top_box)
                self.__widget.attach(self.__top_box, 1, 1, 1, 1)
                self.__widget.attach(self.__labels_box, 1, 0, 2, 1)
                self.__set_album_year(False)
                self.__year_label.set_halign(Gtk.Align.START)
                self.__widget.set_margin_start(MARGIN_SMALL)
                self.__widget.set_margin_end(MARGIN_SMALL)
        elif self.__widget.get_child_at(1, 0) == self.__labels_box:
            self.__widget.remove(self.__labels_box)
            self.__widget.remove(self.__top_box)
            self.__widget.attach(self.__top_box, 2, 0, 1, 1)
            self.__widget.attach(self.__labels_box, 1, 0, 1, 2)
            self.__set_album_year(True)
            self.__year_label.set_halign(Gtk.Align.END)
            self.__widget.set_margin_start(MARGIN)
            self.__widget.set_margin_end(MARGIN)

    def __update_add_button(self):
        """
            Set image as +/-
        """
        albums = App().player.get_albums_for_id(self.__album.id)
        missing = False if albums else True
        for album in albums:
            for track in self.__album.tracks:
                if track.id not in album.track_ids:
                    missing = True
                    break
        if missing:
            self.__add_button.get_image().set_from_icon_name(
                "list-add-symbolic", Gtk.IconSize.BUTTON)
        else:
            self.__add_button.get_image().set_from_icon_name(
                "list-remove-symbolic", Gtk.IconSize.BUTTON)

    def __set_internal_size(self):
        """
            Set content size based on available width
        """
        classes = ["text-medium", "text-large", "text-x-large"]
        # Text size
        for label in [self.__title_label,
                      self.__artist_label,
                      self.__year_label,
                      self.__duration_label]:
            context = label.get_style_context()
            for c in classes:
                context.remove_class(c)

        if App().window.folded:
            art_size = ArtSize.MEDIUM
            icon_size = Gtk.IconSize.BUTTON
            cls_title = "text-medium"
            cls_others = "text-medium"
        elif not self.view_type & ViewType.OVERLAY:
            art_size = ArtSize.BANNER
            icon_size = Gtk.IconSize.LARGE_TOOLBAR
            cls_title = "text-large"
            cls_others = "text-large"
        else:
            art_size = ArtSize.BANNER
            icon_size = Gtk.IconSize.LARGE_TOOLBAR
            cls_title = "text-x-large"
            cls_others = "text-large"
        self.__title_label.get_style_context().add_class(cls_title)
        self.__artist_label.get_style_context().add_class(cls_title)
        self.__year_label.get_style_context().add_class(cls_others)
        self.__duration_label.get_style_context().add_class(cls_others)

        self.__rating_widget.set_icon_size(icon_size)
        self.__loved_widget.set_icon_size(icon_size)
        self.__cover_widget.set_art_size(art_size)

    def __on_cover_populated(self, widget):
        """
            Pass signal
            @param widget as Gtk.Widget
        """
        emit_signal(self, "populated")

    def __on_menu_button_clicked(self, button):
        """
            Show album menu
            @param button as Gtk.Button
        """
        from lollypop.widgets_menu import MenuBuilder
        from lollypop.menu_objects import AlbumMenu
        from lollypop.menu_artwork import AlbumArtworkMenu
        menu = AlbumMenu(self.__album,
                         self.__storage_type,
                         self.view_type)
        menu_widget = MenuBuilder(menu)
        menu_widget.show()
        if App().window.folded:
            menu_ext = AlbumArtworkMenu(self.__album, self.view_type, True)
            menu_ext.connect("hidden", self.__close_artwork_menu)
            menu_ext.show()
            menu_widget.add_widget(menu_ext, -2)
        popup_widget(menu_widget, button, None, None, button)

    def __on_play_button_clicked(self, button):
        """
            Play album
           @param button as Gtk.Button
        """
        selected = button.get_state_flags() & Gtk.StateFlags.SELECTED
        App().player.play_album(self.__album.clone(selected))

    def __on_add_button_clicked(self, button):
        """
            Add/Remove album
           @param button as Gtk.Button
        """
        selected = button.get_state_flags() & Gtk.StateFlags.SELECTED
        add = self.__add_button.get_image().get_icon_name()[0] ==\
            "list-add-symbolic"
        albums = App().player.get_albums_for_id(self.__album.id)
        if albums:
            if add:
                for track in self.__album.tracks:
                    albums[0].append_track(track)
                emit_signal(App().player, "playback-updated", albums[0])
            else:
                for album in albums:
                    for track in self.__album.tracks:
                        album.remove_track(track)
                        if album.id is None:
                            App().player.remove_album(album)
                            break
                emit_signal(App().player, "playback-updated", album)
        else:
            App().player.add_album(self.__album.clone(selected))

    def __on_year_clicked(self, label):
        """
            Show year view
            @param label as LabelWidget
        """
        App().window.container.show_view([Type.YEARS], [self.__album.year])

    def __on_artist_clicked(self, label):
        """
            Show artist view
            @param label as LabelWidget
        """
        App().window.container.show_view([Type.ARTISTS],
                                         self.__album.artist_ids)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import GLib, Gtk

from lollypop.define import ArtSize
from lollypop.helper_size_allocation import SizeAllocationHelper


class HorizontalScrollingHelper(SizeAllocationHelper):
    """
        Manage 2 button and scroll a scrolled window
    """

    def __init__(self):
        """
            Init helper
        """
        SizeAllocationHelper.__init__(self)
        self.__adjustment = self.scrolled.get_hadjustment()
        self.__adjustment.connect("value-changed", self.update_buttons)
        self._backward_button.connect("clicked",
                                      self.__on_backward_button_clicked)
        self._forward_button.connect("clicked",
                                     self.__on_forward_button_clicked)
        self.scrolled.get_hscrollbar().hide()
        self.scrolled.set_policy(Gtk.PolicyType.AUTOMATIC,
                                 Gtk.PolicyType.NEVER)

    def __del__(self):
        """
            Remove ref cycle
        """
        self.__adjustment = None

    def update_buttons(self, *ignore):
        """
            Update buttons state
        """
        value = self.scrolled.get_allocated_width()
        self._backward_button.set_sensitive(self.__adjustment.get_value() !=
                                            self.__adjustment.get_lower())
        self._forward_button.set_sensitive(self.__adjustment.get_value() !=
                                           self.__adjustment.get_upper() -
                                           value)

#######################
# PRIVATE             #
#######################
    def _handle_width_allocate(self, allocation):
        """
            @param allocation as Gtk.Allocation
            @return True if allocation is valid
        """
        if SizeAllocationHelper._handle_width_allocate(self, allocation):
            self.update_buttons()

    def __smooth_scrolling(self, value, direction):
        """
            Emulate smooth scrolling
        """
        if value > 0:
            value -= 1
            current = self.__adjustment.get_value()
            if direction == Gtk.DirectionType.LEFT:
                self.__adjustment.set_value(current - 1)
            else:
                self.__adjustment.set_value(current + 1)
            if value % 10:
                GLib.idle_add(self.__smooth_scrolling, value, direction)
            else:
                GLib.timeout_add(1, self.__smooth_scrolling, value, direction)
        else:
            self.update_buttons()
            self.__adjustment.connect("value-changed", self.update_buttons)

    def __on_backward_button_clicked(self, backward_button):
        """
            Scroll left
            @param backward_button as Gtk.Button
        """
        self.__adjustment.disconnect_by_func(self.update_buttons)
        backward_button.set_sensitive(False)
        value = self.scrolled.get_allocated_width() - ArtSize.BIG
        self.__smooth_scrolling(value, Gtk.DirectionType.LEFT)

    def __on_forward_button_clicked(self, forward_button):
        """
            Scroll right
            @param forward_button as Gtk.Button
        """
        self.__adjustment.disconnect_by_func(self.update_buttons)
        forward_button.set_sensitive(False)
        value = self.scrolled.get_allocated_width() - ArtSize.BIG
        self.__smooth_scrolling(value, Gtk.DirectionType.RIGHT)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import GLib, GObject

from time import time

from lollypop.define import LoadingState, App
from lollypop.logger import Logger
from lollypop.view import View
from lollypop.utils import emit_signal


class LazyLoadingView(View):
    """
        Lazy loading for view
    """

    __gsignals__ = {
        # View has been populated/depopulated, children are not populated
        "initialized": (GObject.SignalFlags.RUN_FIRST, None, ()),
        # All children are populated
        "populated": (GObject.SignalFlags.RUN_FIRST, None, ()),
    }

    def __init__(self, storage_type, view_type):
        """
            Init lazy loading
            @param storage_type as StorageType
            @param view_type as ViewType
        """
        View.__init__(self, storage_type, view_type)
        self.__loading_state = LoadingState.NONE
        self.__lazy_queue = []
        self.__priority_queue = []
        self.__scroll_timeout_id = None
        self.__start_time = time()

    def populate(self, values):
        """
            Populate view with values
            @param values as [object]
        """
        self.__add_values(values)

    def pause(self):
        """
            Pause loading
        """
        self.__loading_state = LoadingState.ABORTED
        if self.__scroll_timeout_id is not None:
            GLib.source_remove(self.__scroll_timeout_id)
            self.__scroll_timeout_id = None
        View.stop(self)

    def stop(self):
        """
            Stop loading and clear queue
        """
        self.__loading_state = LoadingState.FINISHED
        if self.__scroll_timeout_id is not None:
            GLib.source_remove(self.__scroll_timeout_id)
            self.__scroll_timeout_id = None
        self.__lazy_queue = []
        self.__priority_queue = []
        View.stop(self)

    def lazy_loading(self):
        """
            Load the view in a lazy way
        """
        self.__loading_state = LoadingState.RUNNING
        self.__lazy_loading()

    def queue_lazy_loading(self, widget):
        """
            Queue widget into lazy loading
            @param widget as Gtk.Widget
        """
        self.__lazy_queue.append(widget)

    def set_scrolled(self, scrolled):
        """
            Add an external scrolled window
            @param scrolled as Gtk.ScrolledWindow
        """
        View.set_scrolled(self, scrolled)
        scrolled.get_vadjustment().connect("value-changed",
                                           self._on_value_changed)

    @property
    def is_populated(self):
        """
            True if populated
            @return bool
        """
        return self.__loading_state == LoadingState.FINISHED

#######################
# PROTECTED           #
#######################
    def _get_child(self, value):
        """
            Get child for value
            @param value as object
            @return object
        """
        return None

    def _on_map(self, widget):
        """
            Restore backup and load
            @param widget as Gtk.Widget
        """
        View._on_map(self, widget)
        if self.__loading_state == LoadingState.ABORTED and self.__lazy_queue:
            self.lazy_loading()

    def _on_value_changed(self, adj):
        """
            Update scroll value and check for lazy queue
            @param adj as Gtk.Adjustment
        """
        View._on_value_changed(self, adj)
        if not self.__lazy_queue:
            return False
        if self.__scroll_timeout_id is not None:
            GLib.source_remove(self.__scroll_timeout_id)
        self.__scroll_timeout_id = GLib.timeout_add(200, self.__lazy_or_not)

    def _on_populated(self, widget):
        """
            Add another album/disc
            @param widget as AlbumWidget/TracksView
        """
        if self.__loading_state != LoadingState.RUNNING:
            return
        if not widget.is_populated:
            widget.populate()
        else:
            # Do not remove this idle_add()
            # RecursionError: maximum recursion depth exceeded
            GLib.idle_add(self.__lazy_loading)

#######################
# PRIVATE             #
#######################
    def __lazy_loading(self):
        """
            Load the view in a lazy way
        """
        widget = None
        if self.__priority_queue:
            widget = self.__priority_queue.pop(0)
            self.__lazy_queue.remove(widget)
        elif self.__lazy_queue:
            widget = self.__lazy_queue.pop(0)

        if widget is not None:
            widget.connect("populated", self._on_populated)
            widget.populate()
        else:
            self.__loading_state = LoadingState.FINISHED
            emit_signal(self, "populated")
            # Apply filtering
            if App().window.container.type_ahead.get_reveal_child():
                text = App().window.container.type_ahead.entry.get_text()
                if text:
                    self.search_for_child(text)
                else:
                    GLib.idle_add(
                        App().window.container.type_ahead.entry.grab_focus)
            Logger.debug("LazyLoadingView::lazy_loading(): %s",
                         time() - self.__start_time)

    def __add_values(self, values):
        """
            Add widget from values
            @param values as []
        """
        if values:
            value = values.pop(0)
            child = self._get_child(value)
            if child is not None:
                self.__lazy_queue.append(child)
            GLib.idle_add(self.__add_values, values)
        elif self.__loading_state != LoadingState.RUNNING:
            self.__loading_state = LoadingState.RUNNING
            emit_signal(self, "initialized")
            self.__lazy_loading()

    def __is_visible(self, widget):
        """
            Is widget visible in scrolled
            @param widget as Gtk.Widget
        """
        widget_alloc = widget.get_allocation()
        scrolled_alloc = self.scrolled.get_allocation()
        try:
            (x, y) = widget.translate_coordinates(self.scrolled, 0, 0)
            return (y > -widget_alloc.height or y >= 0) and\
                y < scrolled_alloc.height
        except:
            return True

    def __lazy_or_not(self):
        """
            Add visible widgets to lazy queue
        """
        self.__scroll_timeout_id = None
        if self.__loading_state == LoadingState.RUNNING:
            self.__priority_queue = []
            for child in self.__lazy_queue:
                if self.__is_visible(child):
                    self.__priority_queue.append(child)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import GLib, Gtk, Pango, GObject

from lollypop.define import App, ArtSize, ViewType, ArtBehaviour
from lollypop.define import MARGIN, MARGIN_MEDIUM
from lollypop.utils import on_query_tooltip, emit_signal


class AlbumSimpleWidget(Gtk.FlowBoxChild):
    """
        Album widget showing cover, artist and title
    """
    __gsignals__ = {
        "populated": (GObject.SignalFlags.RUN_FIRST, None, ()),
    }

    def __init__(self, album, genre_ids, artist_ids, view_type, font_height):
        """
            Init simple album widget
            @param album as Album
            @param genre_ids as [int]
            @param artist_ids as [int]
            @param view_type as ViewType
            @parma font_height as int
        """
        # We do not use Gtk.Builder for speed reasons
        Gtk.FlowBoxChild.__init__(self)
        self.__album = album
        self.__genre_ids = genre_ids
        self.__artist_ids = artist_ids
        self.__artwork = None
        self.__view_type = view_type
        self.__font_height = font_height
        self.set_property("halign", Gtk.Align.CENTER)
        if view_type & ViewType.ALBUM:
            self.set_property("margin", MARGIN_MEDIUM)
        else:
            self.set_property("margin", MARGIN)
        self.update_art_size()

    def do_get_preferred_width(self):
        """
            Return preferred width
            @return (int, int)
        """
        if self.__artwork is None:
            return (0, 0)
        width = Gtk.FlowBoxChild.do_get_preferred_width(self)[0]
        return (width, width)

    def populate(self):
        """
            Populate widget content
        """
        if self.__artwork is None:
            grid = Gtk.Grid()
            grid.set_orientation(Gtk.Orientation.VERTICAL)
            grid.set_row_spacing(MARGIN_MEDIUM)
            self.__label = Gtk.Label.new()
            self.__label.set_justify(Gtk.Justification.CENTER)
            self.__label.set_ellipsize(Pango.EllipsizeMode.END)
            self.__label.set_property("halign", Gtk.Align.CENTER)
            self.__label.set_property("has-tooltip", True)
            self.__label.connect("query-tooltip", on_query_tooltip)
            style_context = self.__label.get_style_context()
            if self.__view_type & ViewType.SMALL:
                style_context.add_class("text-small")
            album_name = GLib.markup_escape_text(self.__album.name)
            if self.__view_type & ViewType.ALBUM:
                self.__label.set_markup(album_name)
            elif self.__view_type & ViewType.ARTIST:
                if self.__album.year and\
                        App().settings.get_value("show-year-below-name"):
                    self.__label.set_markup(
                        "<b>%s</b>\n<span alpha='25000'>%s</span>" % (
                            album_name, self.__album.year))
                else:
                    self.__label.set_markup("<b>%s</b>" % album_name)
            else:
                artist_name = GLib.markup_escape_text(", ".join(
                                                      self.__album.artists))
                self.__label.set_markup(
                    "<b>%s</b>\n<span alpha='50000'>%s</span>" % (album_name,
                                                                  artist_name))
            self.__artwork = Gtk.Image.new()
            grid.add(self.__artwork)
            grid.add(self.__label)
            self.set_artwork()
            self.set_selection()
            self.connect("destroy", self.__on_destroy)
            self.add(grid)
        else:
            self.set_artwork()

    def update_art_size(self):
        """
            Update art size based on current window state
        """
        if self.__view_type & ViewType.SMALL:
            self.__art_size = ArtSize.MEDIUM
        elif App().window.folded:
            self.__art_size = ArtSize.BANNER
        else:
            self.__art_size = ArtSize.BIG
        self.set_size_request(self.__art_size,
                              self.__art_size + self.__font_height * 2)

    def reset_artwork(self):
        """
            Reset widget artwork
        """
        self.update_art_size()
        if self.__artwork is not None:
            self.__artwork.set_from_surface(None)

    def set_artwork(self):
        """
            Set artwork
        """
        if self.__artwork is None:
            return
        if self.__art_size < ArtSize.BIG:
            frame = "small-cover-frame"
        else:
            frame = "cover-frame"
        App().art_helper.set_frame(self.__artwork,
                                   frame,
                                   self.__art_size,
                                   self.__art_size)
        App().art_helper.set_album_artwork(self.__album,
                                           self.__art_size,
                                           self.__art_size,
                                           self.__artwork.get_scale_factor(),
                                           ArtBehaviour.CACHE |
                                           ArtBehaviour.CROP_SQUARE,
                                           self.__on_album_artwork)

    def set_selection(self):
        """
            Hilight widget if currently playing
        """
        if self.__artwork is None:
            return
        selected = self.__album.id == App().player.current_track.album.id
        if selected:
            self.__artwork.set_state_flags(Gtk.StateFlags.SELECTED, False)
        else:
            self.__artwork.unset_state_flags(Gtk.StateFlags.SELECTED)

    @property
    def data(self):
        """
            @return Album
        """
        return self.__album

    @property
    def name(self):
        """
            Get name
            @return str
        """
        if self.__view_type & (ViewType.ALBUM | ViewType.ARTIST):
            return self.__album.name
        else:
            return "%s %s" % (self.__album.name, self.__album.artists)

    @property
    def artwork(self):
        """
            Get album artwork
            @return Gtk.Image
        """
        return self.__artwork

    @property
    def is_populated(self):
        """
            True if album populated
        """
        return True

#######################
# PRIVATE             #
#######################
    def __on_album_artwork(self, surface):
        """
            Set album artwork
            @param surface as str
        """
        if self.__artwork is None:
            return
        if surface is None:
            if self.__art_size == ArtSize.BIG:
                icon_size = Gtk.IconSize.DIALOG
            else:
                icon_size = Gtk.IconSize.DIALOG.DND
            self.__artwork.set_from_icon_name("folder-music-symbolic",
                                              icon_size)
        else:
            self.__artwork.set_from_surface(surface)
        self.show_all()
        emit_signal(self, "populated")

    def __on_destroy(self, widget):
        """
            Destroyed widget
            @param widget as Gtk.Widget
        """
        self.__artwork = None
